# Rajouter le changement du pion en autre piece (transformation obligée en reine pour l'instant
# Rajouter fin du game (king menacé (ou pris?) en fin de tour)

# Rajouter toutes les pièces vivantes et mortes dans un stack en attribut de game
# Etablir le check mate
# Sauvegarder état partie

import copy

from pyalapin.engine.move import Move
from pyalapin.player.player import Player, AIRandomPlayer
from pyalapin.player.ai_player import EasyAIPlayer
from pyalapin.player.my_player import MyPlayer
import pyalapin.engine.material as material


class Color(object):
    GREEN = "\x1b[32m"
    WHITE = "\033[0m"
    RED = "\x1b[31m"


class Cell(object):
    """
    Cell class representing a base element of a board.

    Attributes
    ----------
    x : int
        x coordinate of the cell on the board.
    y : int
        y coordinate of the cell on the board
    piece: material.Piece or None
        Piece that is on the cell (or None if no Piece is on the cell)
    """

    def __init__(self, x, y, piece):
        """Initialization of the cell.

        Parameters
        ----------
        x : int
            x coordinate of the cell on the board.
        y : int
            y coordinate of the cell on the board
        piece: material.Piece or None
            Piece that is on the cell (or None if no Piece is on the cell)
        """
        self.x = x
        self.y = y
        self.piece = piece
        if piece is not None:
            self.piece.x = x
            self.piece.y = y

    def __deepcopy__(self, memodict={}):
        """Method to create an uncorrelated clone of the cell.

        Returns
        -------
        Cell
            Exact copy of self.
        """
        copy_object = Cell(self.x, self.y, copy.deepcopy(self.piece))
        return copy_object

    def set_piece(self, piece):
        """Sets a Piece in the Cell.

        Parameters
        ----------
        piece: material.Piece
            Piece to set up on self.
        """
        self.piece = piece
        if piece is not None:
            self.piece.x = self.x
            self.piece.y = self.y

    def get_piece(self):
        """Method to access the piece on the Cell.

        Returns
        -------
        Piece
            Piece on the self.
        """
        return self.piece

    def get_x(self):
        """Method to acces Cell x coordinate.

        Returns
        -------
        int
            x-axis coordinate of self.
        """
        return self.x

    def get_y(self):
        """Method to acces Cell y coordinate.

        Returns
        -------
        int
            y-axis coordinate of self.
        """
        return self.y

    def is_threatened(
        self, board, threaten_color
    ):  # change threaten_color par #white_threatened
        """
        Method to check if the Cell is threatened by a given color.

        Parameters
        ----------
        board : Board
            Board to which self belongs to.
        threaten_color : str
            Color of that wants to know if cell is threatened by opponent.

        Returns
        -------
        bool
            Whether the celle is threatened or not.
        """
        # One way that could be more efficient would be to keep at every step the list of threatened cell by each piece
        # And update it at each move.

        # Check Knights threatening
        for i, j in [
            (2, 1),
            (-2, 1),
            (2, -1),
            (-2, -1),
            (1, 2),
            (1, -2),
            (-1, 2),
            (-1, -2),
        ]:
            x_to_check = self.x + i
            y_to_check = self.y + j

            if 0 <= x_to_check < 8 and 0 <= y_to_check < 8:
                cell_to_check = board.get_cell(x_to_check, y_to_check)
                piece_to_check = cell_to_check.get_piece()

                if isinstance(piece_to_check, material.Knight):
                    if piece_to_check.is_white() != threaten_color:
                        return True

        # King + Rook + Queen
        # Checking direct surroundings
        for i, j in [(1, 0), (0, -1), (-1, 0), (0, 1)]:
            x_to_check = self.x + i
            y_to_check = self.y + j
            if 0 <= x_to_check < 8 and 0 <= y_to_check < 8:
                cell_to_check = board.get_cell(x_to_check, y_to_check)
                piece_to_check = cell_to_check.get_piece()

                if (
                    isinstance(piece_to_check, material.King)
                    or isinstance(piece_to_check, material.Rook)
                    or isinstance(piece_to_check, material.Queen)
                ):
                    if piece_to_check.is_white() != threaten_color:
                        return True

                elif piece_to_check is None:
                    keep_going = True
                    x_to_check += i
                    y_to_check += j
                    while 0 <= x_to_check < 8 and 0 <= y_to_check < 8 and keep_going:
                        cell_to_check = board.get_cell(x_to_check, y_to_check)
                        piece_to_check = cell_to_check.get_piece()
                        if isinstance(piece_to_check, material.Rook) or isinstance(
                            piece_to_check, material.Queen
                        ):
                            keep_going = False
                            if piece_to_check.is_white() != threaten_color:
                                return True
                        elif piece_to_check is not None:
                            keep_going = False
                        else:
                            x_to_check += i
                            y_to_check += j

        """
        # Rook + Queen
        # Going further
        keep_going = True
        x_to_check = self.x + 2
        y_to_check = self.y
        while x_to_check < 8 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Rook) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check += 1

        keep_going = True
        x_to_check = self.x - 2
        y_to_check = self.y
        while x_to_check >= 0 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Rook) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check -= 1

        keep_going = True
        x_to_check = self.x
        y_to_check = self.y + 2
        while y_to_check < 8 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Rook) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            y_to_check += 1

        keep_going = True
        x_to_check = self.x
        y_to_check = self.y - 2
        while y_to_check >= 0 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Rook) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            y_to_check -= 1
        """

        # King + Queen + Bishop + Pawn
        # Checking direct surroundings
        for i, j in [(1, 1), (1, -1), (-1, 1), (-1, -1)]:
            x_to_check = self.x + i
            y_to_check = self.y + j

            if 0 <= x_to_check < 8 and 0 <= y_to_check < 8:
                cell_to_check = board.get_cell(x_to_check, y_to_check)
                piece_to_check = cell_to_check.get_piece()

                if (
                    isinstance(piece_to_check, material.King)
                    or isinstance(piece_to_check, material.Bishop)
                    or isinstance(piece_to_check, material.Queen)
                ):
                    if piece_to_check.is_white() != threaten_color:
                        return True
                elif (
                    i > 0
                    and threaten_color
                    and isinstance(piece_to_check, material.Pawn)
                ):
                    if piece_to_check.is_white() != threaten_color:
                        return True
                elif (
                    i < 0
                    and not threaten_color
                    and isinstance(piece_to_check, material.Pawn)
                ):
                    if piece_to_check.is_white() != threaten_color:
                        return True

                elif piece_to_check is None:
                    keep_going = True
                    x_to_check += i
                    y_to_check += j
                    while 0 <= x_to_check < 8 and 0 <= y_to_check < 8 and keep_going:
                        cell_to_check = board.get_cell(x_to_check, y_to_check)
                        piece_to_check = cell_to_check.get_piece()

                        if isinstance(piece_to_check, material.Bishop) or isinstance(
                            piece_to_check, material.Queen
                        ):
                            keep_going = False
                            if piece_to_check.is_white() != threaten_color:
                                return True
                        elif piece_to_check is not None:
                            keep_going = False
                        x_to_check += i
                        y_to_check += j

        """
        # Queen + Bishop
        keep_going = True
        x_to_check = self.x + 2
        y_to_check = self.y + 2
        while x_to_check < 8 and y_to_check < 8 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Bishop) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check += 1
            y_to_check += 1

        keep_going = True
        x_to_check = self.x - 2
        y_to_check = self.y + 2
        while x_to_check >= 0 and y_to_check < 8 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Bishop) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check -= 1
            y_to_check += 1

        keep_going = True
        x_to_check = self.x + 2
        y_to_check = self.y - 2
        while x_to_check < 8 and y_to_check >= 0 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Bishop) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check += 1
            y_to_check -= 1

        keep_going = True
        x_to_check = self.x - 2
        y_to_check = self.y - 2
        while x_to_check >= 0 and y_to_check >= 0 and keep_going:
            cell_to_check = board.get_cell(x_to_check, y_to_check)
            piece_to_check = cell_to_check.get_piece()

            if isinstance(piece_to_check, material.Bishop) or isinstance(
                piece_to_check, material.Queen
            ):
                keep_going = False
                if piece_to_check.is_white() != threaten_color:
                    return True
            elif piece_to_check is not None:
                keep_going = False
            x_to_check -= 1
            y_to_check -= 1
        """

        return False


class Board(object):
    """
    Board class representing the chess board.

    Attributes
    ----------
    board : list of Cells
        Represents all cells of a chess board from x coordinates [0, 7] and y coordinates [0, 7]
    white_king : material.King
        King piece of white color.
    black_king : material.King
        King piece of black color.
    all_material: dict
        Dictionnary containing all the pieces on the board, killed and not killed.
    """

    def __init__(self, empty_init=False):
        """Initialization of the board.

        Parameters
        ----------
        empty_init: bool
            True if you want to start from an existing board.
        """
        if not empty_init:
            self.white_king, self.black_king, self.all_material = self._reset_board()

    def deepcopy(self, light=True):
        """Method to create an uncorrelated clone of the board.

        Returns
        -------
        Cell
            Exact copy of self.
        """
        copied_object = Board(empty_init=True)
        board = [[Cell(i, j, None) for j in range(8)] for i in range(8)]
        copied_object.board = board
        if light:
            copied_material = self.light_deep_copy_material()
        else:
            copied_material = self.deep_copy_material()

        assert (
            len(copied_material["black"]["alive"]["king"]) > 0
        ), "Black king is dead ?"
        assert (
            len(copied_material["white"]["alive"]["king"]) > 0
        ), "White king is dead ?"
        copied_object.white_king = copied_material["white"]["alive"]["king"][0]
        copied_object.black_king = copied_material["black"]["alive"]["king"][0]
        copied_object.all_material = copied_material

        for piece_list in copied_material["white"]["alive"].values():
            for piece in piece_list:
                copied_object.get_cell(piece.x, piece.y).set_piece(piece)
        for piece_list in copied_material["black"]["alive"].values():
            for piece in piece_list:
                copied_object.get_cell(piece.x, piece.y).set_piece(piece)

        return copied_object

    def light_deep_copy_material(self):
        """Method to create an uncorrelated clone of all the pieces on the board. Light version
        where only alive pieces are returned.

        Returns
        -------
        dict of Pieces
            Exact copy of self.all_material.
        """
        material = {
            "white": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
            "black": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
        }

        for color in ["white", "black"]:
            for status in ["alive"]:
                for piece_type in ["pawn", "knight", "bishop", "rook", "queen", "king"]:
                    for piece in self.all_material[color][status][piece_type]:
                        material[color][status][piece_type].append(
                            piece.piece_deepcopy()
                        )
        return material

    def deep_copy_material(self):
        """Method to create an uncorrelated clone of all the pieces on the board, killed and not killed.

        Returns
        -------
        dict of Pieces
            Exact copy of self.all_material.
        """
        material = {
            "white": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
            "black": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
        }

        for color in ["white", "black"]:
            for status in ["alive", "killed"]:
                for piece_type in ["pawn", "knight", "bishop", "rook", "queen", "king"]:
                    for piece in self.all_material[color][status][piece_type]:
                        material[color][status][piece_type].append(
                            piece.piece_deepcopy()
                        )
        return material

    def _deepcopy__(self, memodict={}):
        """Method to create an uncorrelated clone of the board.

        Returns
        -------
        Cell
            Exact copy of self.
        """
        copied_object = Board(empty_init=True)
        board = [[Cell(i, j, None) for j in range(8)] for i in range(8)]
        copied_object.board = board
        copied_material = self.deep_copy_material()

        white_king = copied_material["white"]["alive"]["king"][0]
        black_king = copied_material["black"]["alive"]["king"][0]
        copied_object.all_material = copied_material
        copied_object.white_king = white_king
        copied_object.black_king = black_king
        for piece_list in copied_material["white"]["alive"].values():
            for piece in piece_list:
                copied_object.get_cell(piece.x, piece.y).set_piece(piece)
        for piece_list in copied_material["black"]["alive"].values():
            for piece in piece_list:
                copied_object.get_cell(piece.x, piece.y).set_piece(piece)

        return copied_object

    def to_fen(self):
        """Method to generate a fen representation of the current state of the board

        Returns
        -------
        tuple of str
            fen representation and 'KQkq'
        """
        fen = ""
        for line in self.board:
            no_piece_count = 0
            for cell in line:
                piece = cell.get_piece()
                if piece is None:
                    no_piece_count += 1
                else:
                    if no_piece_count > 0:
                        fen += str(no_piece_count)
                        no_piece_count = 0
                    letter = piece.get_str().replace(" ", "")
                    if piece.is_white():
                        letter = letter.lower()
                    fen += letter
            if no_piece_count > 0:
                fen += str(no_piece_count)
            fen += "/"
        return fen[:-1], "KQkq"

    def one_hot_encode(self, white_side=True):
        """Method to create a representation of the board with OneHot encode of the pieces.

        Parameters
        ----------
        white_sied : bool
            Whether we want to represent the board from the White side point of view or not. Point of view sees its pieces
            represented by +1 OneHot and opponent side by -1.

        Returns
        -------
        list of list
            8x8 list representing the board with full zeros list when cell is empty or OneHot representation of the piece on
            the cell otherwise.
        """

        # Dict of OneHot transformations
        material_to_one_hot = {
            "pawn": [1, 0, 0, 0, 0, 0],
            "bishop": [0, 1, 0, 0, 0, 0],
            "knight": [0, 0, 1, 0, 0, 0],
            "rook": [0, 0, 0, 1, 0, 0],
            "queen": [0, 0, 0, 0, 1, 0],
            "king": [0, 0, 0, 0, 0, 1],
        }
        # Iterating over cells and add OneHot representations to the list.
        one_hot_board = []
        for line in self.board:
            one_hot_line = []
            for cell in line:
                piece = cell.get_piece()
                # Empty cell
                if piece is None:
                    one_hot_line.append([0] * 6)
                # Piece on the cell
                else:
                    one_hot_piece = material_to_one_hot[piece.type]
                    # Negative OneHot if opponent side
                    if piece.is_white() != white_side:
                        one_hot_piece = [-1 * val for val in one_hot_piece]
                    one_hot_line.append(one_hot_piece)
            one_hot_board.append(one_hot_line)
        return one_hot_board

    def get_cell(self, x, y):
        """Method to access a cell on the board from its coordinates.

        Parameters
        ----------
        x : int
            x-coordinate of the cell.
        y : int
            y-coordinate of the cell.

        Returns
        -------
        cell
            Cell at coordinates (x, y)
        """
        return self.board[x][y]

    def reset(self):
        """
        Resets the board, all the pieces, everything.
        """
        self.white_king, self.black_king, self.all_material = self._reset_board()

    def create_board_from_string(self, string):
        """
        Method to set up a sepecific board with Pieces on specific Celss from a string.
        """
        raise NotImplementedError

    def _reset_board(self):
        """Method to create the board. Creates Pieces and place them on their original cells.

        Returns
        -------
        material.King
            White King on the board
        material.King
            black King on the board
        dict
            Dictionnary with all the board pieces
        """
        # List of cells
        board = []

        # Dictionnary to access easily the pieces
        pieces = {
            "white": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
            "black": {
                "alive": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
                "killed": {
                    "pawn": [],
                    "knight": [],
                    "bishop": [],
                    "rook": [],
                    "queen": [],
                    "king": [],
                },
            },
        }

        # Initialize white pieces
        white_king = material.King(True, 0, 4)
        pieces["white"]["alive"]["king"].append(white_king)
        black_king = material.King(False, 7, 4)
        pieces["black"]["alive"]["king"].append(black_king)

        w_rook_1 = material.Rook(True, 0, 0)
        w_rook_2 = material.Rook(True, 0, 7)
        pieces["white"]["alive"]["rook"].append(w_rook_1)
        pieces["white"]["alive"]["rook"].append(w_rook_2)

        w_bishop_1 = material.Bishop(True, 0, 2)
        w_bishop_2 = material.Bishop(True, 0, 5)
        pieces["white"]["alive"]["bishop"].append(w_bishop_1)
        pieces["white"]["alive"]["bishop"].append(w_bishop_2)

        w_knight_1 = material.Knight(True, 0, 1)
        w_knight_2 = material.Knight(True, 0, 6)
        pieces["white"]["alive"]["knight"].append(w_knight_1)
        pieces["white"]["alive"]["knight"].append(w_knight_2)

        w_queen = material.Queen(True, 0, 3)
        pieces["white"]["alive"]["queen"].append(w_queen)

        line = [
            Cell(0, 0, w_rook_1),
            Cell(0, 1, w_knight_1),
            Cell(0, 2, w_bishop_1),
            Cell(0, 3, w_queen),
            Cell(0, 4, white_king),
            Cell(0, 5, w_bishop_2),
            Cell(0, 6, w_knight_2),
            Cell(0, 7, w_rook_2),
        ]
        board.append(line)

        line = []
        for i in range(8):
            p = material.Pawn(True, 1, i)
            pieces["white"]["alive"]["pawn"].append(p)
            line.append(Cell(1, i, p))
        board.append(line)

        for i in range(4):
            line = []
            for j in range(8):
                line.append(Cell(i + 2, j, None))
            board.append(line)

        line = []
        for i in range(8):
            p = material.Pawn(False, 6, i)
            pieces["black"]["alive"]["pawn"].append(p)
            line.append(Cell(6, i, p))
        board.append(line)

        # Initialize black pieces
        b_rook_1 = material.Rook(False, 7, 0)
        b_rook_2 = material.Rook(False, 7, 7)
        pieces["black"]["alive"]["rook"].append(b_rook_1)
        pieces["black"]["alive"]["rook"].append(b_rook_2)

        b_bishop_1 = material.Bishop(False, 7, 2)
        b_bishop_2 = material.Bishop(False, 7, 5)
        pieces["black"]["alive"]["bishop"].append(b_bishop_1)
        pieces["black"]["alive"]["bishop"].append(b_bishop_2)

        b_knight_1 = material.Knight(False, 7, 1)
        b_knight_2 = material.Knight(False, 7, 6)
        pieces["black"]["alive"]["knight"].append(b_knight_1)
        pieces["black"]["alive"]["knight"].append(b_knight_2)

        b_queen = material.Queen(False, 7, 3)
        pieces["black"]["alive"]["queen"].append(b_queen)

        line = [
            Cell(7, 0, b_rook_1),
            Cell(7, 1, b_knight_1),
            Cell(7, 2, b_bishop_1),
            Cell(7, 3, b_queen),
            Cell(7, 4, black_king),
            Cell(7, 5, b_bishop_2),
            Cell(7, 6, b_knight_2),
            Cell(7, 7, b_rook_2),
        ]
        board.append(line)

        self.board = board
        return white_king, black_king, pieces

    def move_piece_from_coordinates(self, start_coordinates, end_coordinates):
        """Method to move a piece on the board from start and landing coordinates.

        Parameters
        ----------
        start_coordinates : tuple of int
            (x, y) coordinates of move starting cell
        end_coordinates : tuple of int
            (x, y) coordinates of move landing cell
        """
        start_cell = self.get_cell(start_coordinates[0], start_coordinates[1])
        end_cell = self.get_cell(end_coordinates[0], end_coordinates[1])
        piece_to_move = start_cell.get_piece()
        if piece_to_move is None:
            raise ValueError("Empty cells chosen as moved piece")

        end_cell.set_piece(piece_to_move)
        start_cell.set_piece(None)

    def kill_piece_from_coordinates(self, coordinates):
        """Method to kill a piece from its coordinates on the board.

        Parameters
        ----------
        coordinates : tuple of ints
            (x, y) coordinates of cell on which is the piece to kill
        """
        to_kill_piece = self.get_cell(coordinates[0], coordinates[1]).get_piece()
        to_kill_piece.set_killed()

        color = "white" if to_kill_piece.is_white() else "black"
        self.all_material[color]["alive"][to_kill_piece.type].remove(to_kill_piece)
        self.all_material[color]["killed"][to_kill_piece.type].append(to_kill_piece)

    def transform_pawn(self, coordinates):
        """Method to promote a pawn from its coordinates.

        Parameters
        ----------
        coordinates : tuple of ints
            (x, y) coordinates of the cell on which is Pawn to promote
        promote_into : str
            type of piece to promote the Pawn into. Default is "Queen" can also be "Rool", "Bishop" and "Knigh"
        """
        pawn = self.get_cell(coordinates[0], coordinates[1]).get_piece()
        if not isinstance(pawn, material.Pawn):
            raise ValueError("Transforming piece that is not a Pawn")
        else:
            color = "white" if pawn.is_white() else "black"
            self.all_material[color]["alive"][pawn.type].remove(pawn)

            new_queen = material.Queen(pawn.is_white(), pawn.x, pawn.y)
            self.get_cell(pawn.x, pawn.y).set_piece(new_queen)
            self.all_material[color]["alive"]["queen"].append(new_queen)

    def promote_pawn(self, coordinates, promote_into="Queen"):
        """Method to promote a pawn from its coordinates.

        Parameters
        ----------
        coordinates : tuple of ints
            (x, y) coordinates of the cell on which is Pawn to promote
        promote_into : str
            type of piece to promote the Pawn into. Default is "Queen" can also be "Rool", "Bishop" and "Knigh"
        """
        pawn = self.get_cell(coordinates[0], coordinates[1]).get_piece()
        if not isinstance(pawn, material.Pawn):
            raise ValueError("Transforming piece that is not a Pawn")
        else:
            color = "white" if pawn.is_white() else "black"
            self.all_material[color]["alive"][pawn.type].remove(pawn)

            new_piece = pawn.promote(promote_into=promote_into)
            self.get_cell(pawn.x, pawn.y).set_piece(new_piece)
            self.all_material[color]["alive"]["queen"].append(new_piece)

    def draw(self, printing=True):
        """Method to draw the board as a string and potentially print it in the terminal.

        Parameters
        ----------
        printing : bool
            Whether or not to print it in the terminal
        """
        whole_text = "    |  0  |  1  |  2  |  3  |  4  |  5  |  6  |  7  |"
        boarder_line = "+---+-----+-----+-----+-----+-----+-----+-----+-----+"
        whole_text += "\n"
        whole_text += boarder_line
        for i in range(8):
            current_line = "  " + str(i) + " |"
            for j in range(8):
                cell = self.get_cell(i, j)
                if cell.get_piece() is None:
                    current_line += "     "
                else:
                    current_line += cell.get_piece().draw()
                current_line += "|"
            whole_text += "\n"
            whole_text += current_line
            whole_text += "\n"
            whole_text += boarder_line
        if printing:
            print(whole_text + "\n")
        return whole_text


class ChessGame(object):
    """
    Game class, used to play a chess game, interact with the board and move pieces.

    Attributes
    ----------
    player1 : player.Player
        player object that will be the one to play white pieces. For now, has to be a human player
    ai : bool
        Whether or not to play with AI. Is set to True, AI will play black pieces.
    player2 : player.Player
        player object that will play the black pieces. Can be human or AI player.
    to_player_player: player1 or player2
        Argument pointing to the player that has to play next. Initialized to white pieces player.
    board: Board
        Board object on which to play.
    status: str
        String indicating if the game is still active or if there is mat or pat.
    played_moves: list
        List storing all the played move during the game.
    automatic draw: bool
        Whether to draw the board in the terminal at each round.
    save_pgn: bool
        Whether to keep track of the moves with PGN.
    history: list of str
        PGN representation of the past move of the game.
    """

    game_status = []

    def __init__(
        self, player1=None, player2=None, automatic_draw=True, ai=False, save_pgn=False
    ):
        """Initialization of the cell.

        Parameters
        ----------
        automatic_draw : bool
            Whether to draw the board in the terminal at each round.
        ai : bool
            Whether or not to play with AI. Is set to True, AI will play black pieces.
        """

        # If ai = True and both players are None, AI plays by default black pieces
        if player2 is None:
            if ai:
                self.player2 = EasyAIPlayer(False)
            else:
                self.player2 = Player(False)

            if player1 is None:
                self.player1 = Player(True)
            else:
                self.player1 = player1

        elif player1 is None:
            if ai:
                self.player1 = EasyAIPlayer(True)
            else:
                self.player1 = Player(True)
            self.player2 = player2
        else:
            self.player1 = player1
            self.player2 = player2

        self.ai = ai
        self.to_play_player = self.player1

        self.board = Board()
        self.status = "ACTIVE"
        self.played_moves = []
        self.save_pgn = save_pgn
        if self.save_pgn:
            self.history = []

        self.automatic_draw = automatic_draw

    def reset_game(self):
        """Method to reset the game. Recreates the borad, the pieces and restarts the game."""
        self.board.reset()
        self.played_moves = []
        self.history = []
        self.to_play_player = self.player1

    def to_fen(self):
        """
        Writes the board in fen.

        Returns
        -------
        str
            fen representation of the board.
        """
        pieces, castling = self.board.to_fen()
        color_playing = "w" if self.to_play_player.is_white_side() else "b"
        return pieces + " " + color_playing + " " + castling + " - 0 1"

    def is_finished(self):
        """
        Method to know if the game is still active or finished (i.e. pat or mat)

        Returns
        -------
        bool
            Whether the game is finished or not.
        """
        return self.status != "ACTIVE"

    def move_from_coordinates(self, player, start_x, start_y, end_x, end_y, extras={}):
        """
        Method to move a piece on the board from its coordinates. Creates the Move object from the coordinates and
        calls the .move() method.

        Parameters
        ----------
        player: player.Player
            player that wants to move a piece
        start_x: int
            x-coordinate of the piece to move
        start_y: int
            y-coordinate of the piece to move
        end_x: int
            x-coordinate of the cell to move the piece to
        end_y: int
            x-coordinate of the cell to move the piece to
        extras: dict
            Dictionnary used to add additional data such as which type a piece a Pawn should be promoted to
            if it reaches the other side of the board.

        Returns
        -------
        self.move()
            Method move of self.
        """
        # Get the cells from the coordinates
        start_cell = self.board.get_cell(start_x, start_y)
        end_cell = self.board.get_cell(end_x, end_y)

        # Create the Move object
        move = Move(player, self.board, start_cell, end_cell, extras=extras)

        # Move
        return self.move(move, player)

    def draw_board(self):
        """
        Draw the game's borad as a string in the terminal.
        """
        return self.board.draw()

    def can_player_move(self, player):
        """
        Methods that verifies if a player can still move at least one piece.

        Parameters
        ----------
        player: player.Player
            player we want to check can still move at least one Piece.

        Returns
        -------
        bool
            Whether player can still move at least a piece.
        """
        # Checks all cells
        for i in range(8):
            for j in range(8):
                # Checks Pieces on cells
                selected_piece = self.board.get_cell(i, j).get_piece()
                if selected_piece is not None:
                    # Checks color of pices
                    if selected_piece.is_white() == player.is_white_side():
                        # Checks if piece can move
                        possible_moves = selected_piece.get_potential_moves(i, j)

                        # Verifies if the move is authorized
                        for k in range(len(possible_moves)):
                            selected_move = possible_moves[k]
                            selected_move = Move(
                                player,
                                self.board,
                                self.board.get_cell(i, j),
                                self.board.get_cell(selected_move[0], selected_move[1]),
                            )
                            verified_move = selected_move.is_possible_move()

                            if verified_move:
                                return True
        return False

    def check_pat_mat(self, player):
        """
        Method to check if a player is in PAT or MAT situation.

        Parameters
        ----------
        player: player.Player
            player that needs to be checked

        Returns
        -------
        int
            0 if the player can move, 1 if PAT, 2 if MAT
        """
        can_player_move = self.can_player_move(player)

        if can_player_move:
            return 0
        # If player cannot move any piece
        else:
            if player.is_white_side():
                king = self.board.white_king
            else:
                king = self.board.black_king
            # If King is threatened, is mat otherwise, is pat
            is_mat = self.board.get_cell(king.x, king.y).is_threatened(
                self.board, not player.is_white_side
            )
            if is_mat:
                return 2
            else:
                return 1

    def move(self, move, player):
        """
        Method to move a piece on the board from player and move objects.

        Parameters
        ----------
        move: move.Move
            move object ready to move a piece.
        player: player.Player
            player that wants to move a piece.

        Returns
        -------
        bool
            Whether the move has happened or not (if not means that it has been blocked by a rule). & Whether the
            game keeps going. (Actually False when not moving should be True)
        int or str
            Status of the game: 0 nothing has happened or no winner, winner otherwise
        """
        moved_piece = move.moved_piece

        # List of checks
        # To change if castling or en passant move
        if moved_piece is None:
            return False, 0
        assert moved_piece is not None

        # Check that right player is playing
        if player != self.to_play_player:
            return False, 0
        assert player == self.to_play_player

        # Check that the move is authorized
        allowed_move = move.is_possible_move()
        if not allowed_move:
            return False, 0
        elif moved_piece.is_white() != player.is_white_side():
            return False, 0
        else:
            assert moved_piece.is_white() == player.is_white_side()
            # Actually move pieces
            move.move_pieces()

        # Store move
        self.played_moves.append(move)
        if self.save_pgn:
            self.history.append(move.to_pgn())

        # Change player
        if self.to_play_player == self.player1:
            self.to_play_player = self.player2
        else:
            self.to_play_player = self.player1

        # Draw
        if self.automatic_draw:
            self.board.draw()

        # Check status of Kings
        if self.board.white_king.is_killed():
            print("END OF THE GAME, BLACK HAS WON")
            return False, "black"
        elif self.board.black_king.is_killed():
            print("END OF THE GAME, WHITE HAS WON")
            return False, "white"

        # Checking for PAT & MAT
        check_status, winner = self.update_status()

        return check_status, winner

    def update_status(self):
        """
        Checks the status of the game (on going, pat, mat) and returns it.

        Returns
        -------
        bool
            Whether the game keeps going or not.
        int or str
            Status of the game: 0 nothing has happened or no winner, winner otherwise

        """
        game_status = self.check_pat_mat(self.player1)
        # Pat
        if game_status == 1:
            return False, "black&white"
        # Mat
        elif game_status == 2:
            return False, "black"
        else:
            game_status = self.check_pat_mat(self.player2)
            # Pat
            if game_status == 1:
                return False, "black&white"
            # Mat
            elif game_status == 2:
                return False, "white"
            # Keeps going
            else:
                return True, ""

    def save(self, directory="debug_files"):
        """
        Method to save the state of the game as matplotlib figure.
        Uses a str representation of the game moves as figure title.

        Parameters
        ----------
        directory: str
            directory in which to save the figure.
        """
        draw_text = self.draw_board()
        draw_text = draw_text.replace("\x1b[32m", "")
        draw_text = draw_text.replace("\033[0m", "")
        draw_text = draw_text.replace("\x1b[31m", "")
        import os
        import matplotlib.pyplot as plt

        """
        plt.rc("figure", figsize=(12, 7))
        plt.text(
            0.01, 0.05, str(draw_text), {"fontsize": 10}, fontproperties="monospace"
        )
        plt.axis("off")
        plt.tight_layout()
        plt.savefig(os.path.join(directory, str(len(self.played_moves)) + ".png"))
        """
        with open(
            os.path.join(directory, str(len(self.played_moves)) + ".txt"), "w"
        ) as f:
            f.writelines(draw_text)

    def to_pgn(self):
        assert self.save_pgn
        pgn = ""
        for i in range(len(self.history)):
            if i % 2 == 0:
                pgn += f"{int(i/2)+1}. "
            pgn += self.history[i]
            pgn += " "

        return pgn[:-1]



import copy
import pickle

import pyalapin.engine.material as material


class Move(object):
    """Base class for material movement.

    Implements various checks and is able to operate a piece movement along additional actions such as taking an
    adversary piece, castling, en passant.

    Attributes
    ----------
    player : player.Player
        Which player wants to play the move.
    board : Engine.Board
        board of the game on which the move is operated.
    start : engine.Cell
        Current cell on which the moved piece is.
    end: engine.Cell
        Cell on which the piece is moved to.
    moved_piece: material.Piece
        Piece of the board that is moved.
    killed_piece: [None or material.Piece]
        If a piece is killed during the move (meaning that there is a piece on cell end) otherwise is None.
    is_castling: bool
        Whether the move it castling.
    complementary_castling: [None or material.Piece]
        If is_castling, holds additional information => To be moved in extras ?
    is_en_passant: bool
        Whether or not the move is taking a Pawn 'En Passant'
    extras: dict
        Used to introduce additional information such as piece for promotion.

    """

    def __init__(self, player, board, start, end, extras={}):
        """Initialization of the piece.

        Parameters
        ----------
        player : player.Player
            Which player wants to play the move.
        board : Engine.Board
            board of the game on which the move is operated.
        start : engine.Cell
            Current cell on which the moved piece is.
        end: engine.Cell
            Cell on which the piece is moved to.
        extras: dict
            Used to introduce additional information such as piece for promotion.

        """
        self.player = player
        self.board = board
        self.start = start  # Should we take only starting coordinates ?
        self.end = end  # Should we take only landing coordinates ?
        self.extras = extras

        self.moved_piece = start.get_piece()  # Check that moved_piece is not None ?

        self.killed_piece = self.end.get_piece()
        self.is_castling = False
        self.complementary_castling = None
        self.en_passant = False

    def deepcopy(self):
        """Method to create an uncorrelated clone of the move.

        Returns
        -------
        Move
            Exact copy of self.

        """
        # Rethink what needs to be copied and what needs not.
        copied_board = self.board.deepcopy()
        copied_move = Move(
            self.player,
            copied_board,
            copied_board.get_cell(self.start.x, self.start.y),
            copied_board.get_cell(self.end.x, self.end.y),
        )
        copied_move.is_castling = self.is_castling
        copied_move.complementary_castling = self.complementary_castling
        copied_move.en_passant = self.en_passant
        return copied_move

    def _set_moved_attribute(self):
        """Method to set the 'has_moved' attributes for the pieces that need to check whether
        they have already moved or not.
        Also, if the piece is a Pawn, changes the last_move_is_double attribute so that it can
        be taken 'En Passant' if this is the case.
        """
        if hasattr(
            self.moved_piece, "has_moved"
        ):  # Check if the moved piece has the attribute
            self.moved_piece.has_moved = True

        if hasattr(self.moved_piece, "last_move_is_double"):  # Check if it is a Pawn
            if (
                abs(self.start.get_x() - self.end.get_x()) > 1
            ):  # If the move is a double forward
                self.moved_piece.last_move_is_double = True
            else:
                self.moved_piece.last_move_is_double = False

    def to_pgn(self):
        """
        Method to return the PGN representation of the move.

        Returns
        -------
        str:
            pgn representation of the move
        """
        rows = ["a", "b", "c", "d", "e", "f", "g", "h"]
        start = f"{rows[self.start.y]}{self.start.x + 1}"
        end = f"{rows[self.end.y]}{self.end.x + 1}"
        piece = self.moved_piece.get_str().replace(" ", "")
        if isinstance(self.moved_piece, material.Pawn):
            piece = ""
        if self.killed_piece is not None:
            start += "x"
        elif self.is_castling:
            if (self.moved_piece.is_white() and self.end.y == 1) or (
                not self.moved_piece.is_white() and self.end.y == 6
            ):
                piece = "O-O-O"
                start = ""
                end = ""
        king = (
            self.board.white_king
            if not self.player.white_side
            else self.board.black_king
        )
        if self.board.get_cell(king.x, king.y).is_threatened(
            board=self.board, threaten_color=not self.player.white_side
        ):
            end += "+"
        print(f"{piece}{start}{end}")
        return f"{piece}{start}{end}"

    def _set_castling_done(self):
        """
        If self is a castling move, then when it is done this function sets the castling_done attribute
        of the concerned King to True, so that it cannot use castling twice.
        """
        assert isinstance(self.moved_piece, material.King)
        self.moved_piece.castling_done = True

    def _is_castling(self):
        """
        Checks if the current move is castling. In particular verifies that:
            - the moved piece is a King
            - the moved King has not already used its castling
            - the King has not already moved
            - the comparnion Rook is on the right Cell
            - the companion Rook has not already moved
            - the Cells between the King and the Rook are empty
            - the King landing Cell is not threatened by adversary Pieces
        Also if it is a castling move sets in self.extra the Rook that needs to be moved along the King

        Returns
        -------
        bool
            Whether self is castling or not.
        """
        if not isinstance(self.moved_piece, material.King):
            ###print("not castling becasuse king not moved")
            return False
        elif self.moved_piece.castling_done or self.moved_piece.has_moved:
            ###print(self.moved_piece.castling_done)
            ###print(self.moved_piece.has_moved)
            return False

        else:
            if self.end.y == 6:  # Castling on the right
                rook_to_move = self.board.get_cell(self.start.x, 7).get_piece()
                if not isinstance(rook_to_move, material.Rook):
                    return False
                elif rook_to_move.has_moved:
                    return False
                else:
                    rook_starting_coordinates = (self.start.x, 7)
                    rook_ending_coordinates = (self.start.x, 5)
                    must_be_empty_cells = [
                        self.board.get_cell(self.start.x, 5),
                        self.board.get_cell(self.start.x, 6),
                    ]
                    must_not_be_threatened_cells = [
                        self.board.get_cell(self.start.x, 4),
                        self.board.get_cell(self.start.x, 5),
                        self.board.get_cell(self.start.x, 6),
                    ]

            elif self.end.y == 2:  # Castling on the left
                rook_to_move = self.board.get_cell(self.start.x, 0).get_piece()
                if not isinstance(rook_to_move, material.Rook):
                    return False
                elif rook_to_move.has_moved:
                    return False
                else:
                    rook_starting_coordinates = (self.start.x, 0)
                    rook_ending_coordinates = (self.start.x, 3)
                    must_be_empty_cells = [
                        self.board.get_cell(self.start.x, 1),
                        self.board.get_cell(self.start.x, 2),
                        self.board.get_cell(self.start.x, 3),
                    ]
                    must_not_be_threatened_cells = [
                        self.board.get_cell(self.start.x, 2),
                        self.board.get_cell(self.start.x, 3),
                        self.board.get_cell(self.start.x, 4),
                    ]
            else:
                ###print('king did not move to a castling position')
                return False

            # Verifying that the cells between rook and king are empty and that starting
            # and landing cells for the king are not threatened.
            empty_cells_check = True
            not_threatened_cells = True
            for cll in must_be_empty_cells:
                if cll.get_piece() is not None:
                    empty_cells_check = False
            for cll in must_not_be_threatened_cells:
                if cll.is_threatened(self.board, self.moved_piece.is_white()):
                    not_threatened_cells = False

            # Verifies that both conditions are met
            conditions_to_castling = [empty_cells_check, not_threatened_cells]
            if all(conditions_to_castling):
                self.complementary_castling = (  # To store in self.extras :)
                    rook_to_move,
                    self.board.get_cell(
                        rook_starting_coordinates[0], rook_starting_coordinates[1]
                    ),
                    self.board.get_cell(
                        rook_ending_coordinates[0], rook_ending_coordinates[1]
                    ),
                )
                self.extras["complementary_castling"] = (
                    rook_to_move,
                    self.board.get_cell(
                        rook_starting_coordinates[0], rook_starting_coordinates[1]
                    ),
                    self.board.get_cell(
                        rook_ending_coordinates[0], rook_ending_coordinates[1]
                    ),
                )
                return True

            else:
                ###print('Conditions for castling:')
                ###print('Rook has not moved:', rook_to_move.has_moved)
                ###print('Cells in between empty:', empty_cells_check)
                ###print('Cells in between not threatened:', not_threatened_cells)
                return False

    def _is_en_passant(self):
        """
        Checks if the current move is an 'En passant' capture. In particular verifies that:
            - the moved piece is a Pawn
            - the movement of the pices is in diagonal
            - the move crosses another Pawn
            - this crossed Pawn is of different color
            - this crossed Pawn last move is a double forward advance
        Also if it is an 'En passant' capture sets the crossed Pawn as killed_piece
        Returns
        -------
        bool
            Whether self is an 'En passant' capture or not.
        """
        if isinstance(
            self.moved_piece, material.Pawn
        ):  # Only a Pawn can take 'En Passant'
            dx = self.start.get_x() - self.end.get_x()
            dy = self.start.get_y() - self.end.get_y()
            # Needs the movement to be in diagonal and that no piece is on the landing Cell
            if dy == 0 or self.killed_piece is not None:
                return False

            else:
                # Retrieving crossed Piece
                crossed_cell = self.board.get_cell(self.start.get_x(), self.end.get_y())
                crossed_piece = crossed_cell.get_piece()
                # Verifying the crossed Piece is a Pawn
                if isinstance(crossed_piece, material.Pawn):
                    # Verifying color and last move of crossed_piece
                    if (
                        crossed_piece.last_move_is_double
                        and crossed_piece.is_white() != self.moved_piece.is_white()
                    ):
                        # Revoir comment on update cet attribut last_move_is_double
                        self.killed_piece = crossed_piece
                        self.en_passant = True
                        return True
                    else:
                        return False
                else:
                    return False
        else:
            return False

    def _is_pawn_promotion(self):
        """
        Checks if the current move is should ends up with a Pawn promotion, meaning that:
            - the moved piece is a Pawn
            - the Pawn reaches the other side of the board according to its color
        If the Piece it should be promote into is not specified in self.extras then sets Queen as promotion.

        Returns
        -------
        bool
            Whether self should ends up by a Pawn promotion or not.
        """
        # Checks the piece
        if not isinstance(self.moved_piece, material.Pawn):
            return False
        else:
            # Checks if the Pawn has reached the other side of the board.
            if self.end.get_x() == 7 and self.moved_piece.is_white():  # White Piece
                # Standard is to promote into a Queen if not specified
                self.promote_into = self.extras.get("promote_into", "queen")
                return True

            elif (
                self.end.get_x() == 0 and not self.moved_piece.is_white()
            ):  # Black Piece
                # Standard is to promote into a Queen if not specified
                self.promote_into = self.extras.get("promote_into", "queen")
                return True
            else:
                return False

    def _promote_pawn(self):
        """
        Organizes the Pawn promotion.
        """
        coordinates = (self.end.get_x(), self.end.get_y())
        self.board.promote_pawn(coordinates=coordinates, promote_into=self.promote_into)

    def move_pieces(self):
        """
        Effectively moves pieces on board
        """
        # Do everything from coordinates so that only board needs to be copied in self.deepcopy() ?

        # Kills Piece on landing Cell if needed.
        if self.killed_piece is not None:
            self.board.kill_piece_from_coordinates((self.end.x, self.end.y))
        # Moves Piece on the Board
        self.board.move_piece_from_coordinates(
            (self.start.x, self.start.y), (self.end.x, self.end.y)
        )

        # Executes castling if needed
        if self.complementary_castling is not None and self.is_castling:
            castling_rook, rook_start, rook_end = self.complementary_castling
            self.board.move_piece_from_coordinates(
                (rook_start.x, rook_start.y), (rook_end.x, rook_end.y)
            )

            # Sets castling to done
            self._set_castling_done()

        # Promotes Pawn if needed
        if self._is_pawn_promotion():
            self._promote_pawn()

        # Sets the different movement related attributes of Pieces
        self._set_moved_attribute()

    def is_possible_move(self, check_chess=True):
        # REFONDRE, particulièrement, faire en sorte qu'on ne vérifie chaque condition qu'une seule fois
        # Why castling is checked here ?
        """
        Checks if move is possible. In particular checks that:
            - Landing Cell is different than current Cell
            - Checks the color of the potential Piece on the landing Cell
            - Movement on the Board is authorized according to the Piece type
            - If moved Piece is a King, does not land on a threatened Cell

        If the Piece it should be promote into is not specified in self.extras then sets Queen as promotion.

        Returns
        -------
        bool
            Whether the move is legal or not.
        """
        # Should be kept ?

        dx = self.start.get_x() - self.end.get_x()
        dy = self.start.get_y() - self.end.get_y()

        # If no movement not possible
        if abs(dx) == 0 and abs(dy) == 0:
            return False

        """
        # Check color of player and color of moving piece:
        if self.moved_piece.is_white() != self.player.is_white_side():
            ###print(self.moved_piece, self.moved_piece.is_white(), self.player.is_white_side())
            ###print("Player", self.player, "tries to move adversaries pieces")
            return False
        """

        # check color of receiving piece
        if self.killed_piece is not None:
            if self.moved_piece.is_white() == self.killed_piece.is_white():
                return False
            if self.killed_piece.is_white() == self.player.is_white_side():
                return False

        # good moving for type of piece selected
        is_legal_move = self.moved_piece.can_move(self.board, self)

        # Why here ?
        self.is_castling = self._is_castling()

        if not is_legal_move:
            if not self._is_en_passant() or not self.is_castling:
                return False

        """
        # Check for chess
        def work_future(curr_move, curr_board):
            future_board = copy.deepcopy(curr_board)
            x_start = curr_move.start.get_x()
            y_start = curr_move.start.get_y()
            x_end = curr_move.end.get_x()
            y_end = curr_move.end.get_y()
            future_piece = future_board.get_cell(x_start, y_start).get_piece()

            future_board.get_cell(x_end, y_end).set_piece(future_piece)
            future_board.get_cell(x_start, y_start).set_piece(None)

            if hasattr(future_piece, 'has_moved'):
                future_piece.has_moved = True
            if hasattr(future_piece, 'last_move_is_double'):
                if abs(x_start - x_end) > 1:
                    future_piece.last_move_is_double = True
                else:
                    future_piece.last_move_is_double = False

            # Check for castling
            if curr_move.complementary_castling is not None:
                ###print('Apparently future castling move, working on moving the Rook')
                castling_rook, rook_start, rook_end = curr_move.complementary_castling
                rook_start = future_board.get_cell(rook_start.get_x(), rook_start.get_y())
                rook_end = future_board.get_cell(rook_end.get_x(), rook_end.get_y())
                castling_rook = rook_start.get_piece()

                rook_end.set_piece(castling_rook)
                rook_start.set_piece(None)

            # Check for prise en passant
            if curr_move.complementary_passant is not None:
                ###print('This future move is taking En Passant')
                x_passant = curr_move.complementary_passant.get_x()
                y_passant = curr_move.complementary_passant.get_y()

                en_passant_piece = future_board.get_cell(x_passant, y_passant).get_piece()
                en_passant_piece.set_killed(True)
                future_board.get_cell(x_passant, y_passant).set_piece(None)

            # Checking if future king is chessed
            if curr_move.player.is_white_side():
                king = future_board.white_king
            else:
                king = future_board.black_king

            return future_board.get_cell(king.x, king.y).is_threatened(future_board, king.is_white())
        """
        # Checks if the player's King is threatened after the move.
        if check_chess:
            is_king_threatened_in_future = self._work_future_to_check_chess()
            if is_king_threatened_in_future:
                return False

        return True

    def _work_future_to_check_chess(self):  # Can it be done better ?
        """
        Effectively move the Piece and check if the King is then threatened. In this case, move cannot happen.

        Returns
        -------
        bool
            Whether the player's King is threatened once the move done.
        """

        # Deep Copy everything
        move = self.deepcopy()
        # Move Piece
        move.move_pieces()

        # Select player's King and check if threatened.
        if move.player.is_white_side():
            king = move.board.white_king
        else:
            king = move.board.black_king
        return move.board.get_cell(king.x, king.y).is_threatened(
            move.board, king.is_white()
        )



from abc import abstractmethod

from pyalapin.engine.color import Color


class Piece(object):
    """Base class for the different materials on the board.

    Implements the properties, attributes and functions shared by all the different materials.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized as False

    """

    def __init__(self, white, x, y):
        """Initialization of the piece.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        self.white = white
        self.killed = False

        self.x = x
        self.y = y

    @abstractmethod
    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Piece
            Exact copy of self.

        """
        copied_piece = Piece(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        return copied_piece

    def is_white(self):
        """Method to access the piece color.

        Returns
        -------
        bool
            color of piece.

        """
        return self.white

    def is_killed(self):
        """Method to access the piece status (killed or not).

        Returns
        -------
        bool
            status of piece.

        """
        return self.killed

    def set_killed(self):
        """Sets the piece status to killed."""
        self.killed = True

    @abstractmethod
    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell of piece).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        x_start = start.get_x()
        y_start = start.get_y()
        x_end = end.get_x()
        y_end = end.get_y()

        # If material is already on the landing cell
        if end.get_piece() is not None:
            if end.get_piece().is_white() == self.is_white():
                return False

        if x_start == x_end and y_start == y_end:
            return False
        else:
            if x_start < 0 or x_end < 0:
                return False
            elif x_start > 7 or x_end > 7:
                return False
            elif y_start < 0 or y_end < 0:
                return False
            elif y_start > 7 or y_end > 7:
                return False
            else:
                return True

    def can_move(self, board, move):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """

        is_movement_authorized = self.piece_move_authorized(move.start, move.end)
        return is_movement_authorized

    @abstractmethod
    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        bool
            List of authorized moves
        """
        return None

    @abstractmethod
    def get_threatened_cells_on_board(self, board):
        """
        Method to list which cells are threatened by the piece (authorized movement + conditioned on other pieces on board).

        Parameters
        ----------
        board: Board
            game board self belong to

        Returns
        -------
        list
            List of threatened cells

        """
        return []

    @abstractmethod
    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "     "

    def draw(self):
        """Method to represent the piece as a colored string in order to draw a board.

        Returns
        -------
        str
            Colored representation of the piece within a board.
        """
        value = self.get_str()
        if self.is_white():
            return Color.GREEN + value + Color.WHITE
        else:
            return Color.RED + value + Color.WHITE


class Pawn(Piece):
    """Base class for the pawns pieces

    Implements the properties, attributes and functions specific to pawns.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False
    has_moved: bool
        Whether the piece has already moved during a game or not. Initialized to False
    last_move_is_double: bool
        Whether the piece previous move was a double advance or not. For future En Passant checks. Initialized to False.
    """

    type = "pawn"

    def __init__(self, *args, **kwargs):
        """Initialization of the pawn.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)
        self.has_moved = False  # if the pawn has yet been moved or not to keep ?
        self.last_move_is_double = (
            False  # check for en passant, if last move was a double tap
        )

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Pawn
            Exact copy of self.
        """
        copied_piece = Pawn(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        copied_piece.has_moved = self.has_moved
        copied_piece.last_move_is_double = self.last_move_is_double
        return copied_piece

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        # Check if there is a piece on the landing cell
        if not super().piece_move_authorized(start=start, end=end):
            return False
        if end.get_piece() is not None:
            # check if there is not another piece of same color
            if end.get_piece().is_white() == self.is_white():
                return False

            else:
                # Pawn can only take an adversary piece in diagonal
                dx = end.get_x() - start.get_x()
                dy = end.get_y() - start.get_y()
                if dx == 1 and abs(dy) == 1 and self.is_white():
                    return True
                elif dx == -1 and abs(dy) == 1 and not self.is_white():
                    return True
                else:
                    return False

        # No piece on landing cell, just checking if movement is authorized.
        else:
            dx = end.get_x() - start.get_x()
            dy = end.get_y() - start.get_y()
            if dx == 1 and dy == 0 and self.is_white():
                return True
            elif dx == -1 and dy == 0 and not self.is_white():
                return True

            else:
                # Initial move authorized to be two cells at once. Should check self.has_moved here ?
                if start.get_x() == 1 and dx == 2 and dy == 0 and self.is_white():
                    return True
                elif (
                    start.get_x() == 6 and dx == -2 and dy == 0 and not self.is_white()
                ):
                    return True
                else:
                    return False

    def can_move(self, board, move):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        authorized_move = self.piece_move_authorized(move.start, move.end)

        if not authorized_move:
            """to remove ?"""
            crossed_cell = board.get_cell(move.start.get_x(), move.end.get_y())
            crossed_piece = crossed_cell.get_piece()
            if isinstance(crossed_piece, Pawn):
                if (
                    crossed_piece.last_move_is_double
                    and crossed_piece.is_white() != self.is_white()
                ):
                    # Revoir comment on update cet attribut last_move_is_double
                    authorized_move = True
                    move.complementary_passant = crossed_cell
        else:
            # Checks that no piece (friend or foe) is blocking the cell(s) in front.
            dx = move.end.get_x() - move.start.get_x()

            if dx > 1:
                if (
                    board.get_cell(
                        move.start.get_x() + 1, move.start.get_y()
                    ).get_piece()
                    is not None
                ):
                    return False
            elif dx < -1:
                if (
                    board.get_cell(
                        move.start.get_x() - 1, move.start.get_y()
                    ).get_piece()
                    is not None
                ):
                    return False
        """
        if move.end.get_x() == 7 and self.is_white():
            move.transform_pawn = True
        elif move.end.get_x() == 0 and not self.is_white():
            move.transform_pawn = True
        """
        return authorized_move

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """

        possible_moves = []
        if self.is_white():
            # Front cell
            if x < 7:
                possible_moves.append((x + 1, y))

                # Diagonal cells
                if y - 1 >= 0:
                    possible_moves.append((x + 1, y - 1))
                if y + 1 <= 7:
                    possible_moves.append((x + 1, y + 1))

            # Double front cell
            if x == 1:
                possible_moves.append((x + 2, y))

        # Symmetric for black pawns
        else:
            if x > 0:
                possible_moves.append((x - 1, y))

                if y - 1 >= 0:
                    possible_moves.append((x - 1, y - 1))
                if y + 1 <= 7:
                    possible_moves.append((x - 1, y + 1))
            if x == 6:
                possible_moves.append((x - 2, y))

        return possible_moves

    @abstractmethod
    def get_threatened_cells_on_board(self, board):
        """
        Method to list which cells are threatened by the piece (authorized movement + conditioned on other pieces on board).

        Parameters
        ----------
        board: Board
            game board self belong to

        Returns
        -------
        list
            List of threatened cells

        """

        cells_threatened = []
        if self.is_white():
            if x < 7:
                # Diagonal cells
                if y - 1 >= 0:
                    cells_threatened.append((x + 1, y - 1))
                if y + 1 <= 7:
                    cells_threatened.append((x + 1, y + 1))

        # Symmetric for black pawns
        else:
            if x > 0:
                if y - 1 >= 0:
                    cells_threatened.append((x - 1, y - 1))
                if y + 1 <= 7:
                    cells_threatened.append((x - 1, y + 1))

        return cells_threatened

    def promote(self, promote_into="Queen"):
        """Method to promote a pawn to other material type. Only happens if the pawn reaches the other side of the board.
        The player can choose which type of material he wants its pawn to be promoted into.

        Parameters
        ----------
        promote_into: str
            Type of material to promote the pawn into.

        Returns
        -------
        Piece
            New piece with right promotion and same coordinates of current pawn.
        """
        # Should we verify color and coordinates of Pawn ?
        if self.is_white():
            assert self.x == 7, "Pawn has not reached other side of the board"
        else:
            assert self.x == 0, "Pawn has not reached the other side of the board"

        if promote_into.lower() == "queen":
            return Queen(white=self.is_white(), x=self.x, y=self.y)
        elif promote_into.lower() == "rook":
            rook = Rook(white=self.is_white(), x=self.x, y=self.y)
            rook.has_moved = True
            return rook
        elif promote_into == "knight":
            return Knight(white=self.is_white(), x=self.x, y=self.y)
        elif promote_into == "bishop":
            return Bishop(white=self.is_white(), x=self.x, y=self.y)
        else:
            raise ValueError(
                f"Cannot promote piece into, {promote_into}, piece unknown"
            )

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  P  "


class Bishop(Piece):
    """Base class for the bishop pieces

    Implements the properties, attributes and functions specific to bishops.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False.
    """

    type = "bishop"

    def __init__(self, *args, **kwargs):
        """Initialization of the bishop.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Bishop
            Exact copy of self.
        """
        copied_piece = Bishop(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        return copied_piece

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        if not super().piece_move_authorized(start=start, end=end):
            return False

        # Checking movemement
        dx = end.get_x() - start.get_x()
        dy = end.get_y() - start.get_y()
        if abs(dx) == abs(dy):
            return True
        else:
            return False

    def can_move(self, board, move):
        """Method to verify if a move is authorized within a board.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        # Checking if movement is alright
        authorized_move = self.piece_move_authorized(move.start, move.end)
        if authorized_move:
            dx = move.end.get_x() - move.start.get_x()
            dy = move.end.get_y() - move.start.get_y()

            # Checking that no material is blocking the trajectory
            for i in range(1, abs(dx)):
                x_trajectory = i * int(dx / abs(dx)) + move.start.get_x()
                y_trajectory = i * int(dy / abs(dy)) + move.start.get_y()
                if board.get_cell(x_trajectory, y_trajectory).get_piece() is not None:
                    return False
            return True
        else:
            return False

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """
        possible_moves = []

        # Diagonal 1
        nx = x - 1
        ny = y - 1
        while nx >= 0 and ny >= 0:
            possible_moves.append((nx, ny))
            nx -= 1
            ny -= 1

        # Diagonal 2
        nx = x - 1
        ny = y + 1
        while nx >= 0 and ny <= 7:
            possible_moves.append((nx, ny))
            nx -= 1
            ny += 1

        # Diagonal 3
        nx = x + 1
        ny = y - 1
        while nx <= 7 and ny >= 0:
            possible_moves.append((nx, ny))
            nx += 1
            ny -= 1

        # Diagonal 4
        nx = x + 1
        ny = y + 1
        while nx <= 7 and ny <= 7:
            possible_moves.append((nx, ny))
            nx += 1
            ny += 1

        return possible_moves

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  B  "


class Rook(Piece):
    """Base class for the rook pieces

    Implements the properties, attributes and functions specific to rooks.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False.
    has_moved: bool
        Whether the piece has already moved during a game or not. Initialized to False.
    """

    type = "rook"

    def __init__(self, *args, **kwargs):
        """Initialization of the rook.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)
        self.has_moved = False

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Rook
            Exact copy of self.
        """
        copied_piece = Rook(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        copied_piece.has_moved = self.has_moved
        return copied_piece

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        if not super().piece_move_authorized(start=start, end=end):
            return False

        # Checking movement
        dx = end.get_x() - start.get_x()
        dy = end.get_y() - start.get_y()
        if dx == 0 or dy == 0:
            return True
        else:
            return False

    def can_move(self, board, move):
        """Method to verify if a move is authorized within a board.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """

        # Checking that movement is authorized
        authorized_move = self.piece_move_authorized(move.start, move.end)
        if authorized_move:
            dx = move.end.get_x() - move.start.get_x()
            dy = move.end.get_y() - move.start.get_y()
            # Checking that no material in x axis is blocking the trajectory
            for i in range(1, abs(dx)):
                x_trajectory = i * int(dx / abs(dx)) + move.start.get_x()
                y_trajectory = move.start.get_y()
                if board.get_cell(x_trajectory, y_trajectory).get_piece() is not None:
                    return False

            # Checking that no material in the y axis is blocking the trajectory
            for i in range(1, abs(dy)):
                x_trajectory = move.start.get_x()
                y_trajectory = i * int(dy / abs(dy)) + move.start.get_y()
                if board.get_cell(x_trajectory, y_trajectory).get_piece() is not None:
                    return False
            return True
        else:
            return False

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """
        possible_moves = []

        # X-axis left
        nx = x - 1
        while nx >= 0:
            possible_moves.append((nx, y))
            nx -= 1

        # X-axis right
        ny = y + 1
        while ny <= 7:
            possible_moves.append((x, ny))
            ny += 1

        # Y-axis top
        nx = x + 1
        while nx <= 7:
            possible_moves.append((nx, y))
            nx += 1

        # Y-axis down
        ny = y - 1
        while ny >= 0:
            possible_moves.append((x, ny))
            ny -= 1

        return possible_moves

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  R  "


class Knight(Piece):
    """Base class for the knoght pieces

    Implements the properties, attributes and functions specific to knights.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False.
    """

    type = "knight"

    def __init__(self, *args, **kwargs):
        """Initialization of the knight.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Knight
            Exact copy of self.
        """
        copied_piece = Knight(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        return copied_piece

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        if not super().piece_move_authorized(start=start, end=end):
            return False

        dx = start.get_x() - end.get_x()
        dy = start.get_y() - end.get_y()
        return abs(dx) * abs(dy) == 2

    def can_move(self, board, move):
        """Method to verify if a move is authorized within a board.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        # The knight is jumping, no need to verify blocking material
        return self.piece_move_authorized(move.start, move.end)

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  N  "

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """
        possible_moves = []

        # All difference position that a knight can move to
        combos = [
            (2, 1),
            (1, 2),
            (-2, 1),
            (2, -1),
            (-2, -1),
            (-1, 2),
            (1, -2),
            (-1, -2),
        ]
        for nx, ny in combos:
            if 0 <= nx + x <= 7 and 0 <= ny + y <= 7:
                possible_moves.append((x + nx, y + ny))

        return possible_moves


class Queen(Piece):
    """Base class for the queen pieces

    Implements the properties, attributes and functions specific to queens.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False.
    """

    type = "queen"

    def __init__(self, *args, **kwargs):
        """Initialization of the queen.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        Queen
            Exact copy of self.
        """
        copied_piece = Queen(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        return copied_piece

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """

        if not super().piece_move_authorized(start=start, end=end):
            return False

        dx = end.get_x() - start.get_x()
        dy = end.get_y() - start.get_y()

        return (dx == 0) or (dy == 0) or (abs(dx) == abs(dy))

    def can_move(self, board, move):
        """Method to verify if a move is authorized within a board.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        # Checking if movement is authorized
        authorized_move = self.piece_move_authorized(move.start, move.end)

        # Checking that no material is blocking the trajectory
        if authorized_move:
            dx = move.end.get_x() - move.start.get_x()
            dy = move.end.get_y() - move.start.get_y()

            # Queen going along an axis
            if dx == 0 or dy == 0:
                # Along X-axis
                for i in range(1, abs(dx)):
                    x_trajectory = i * int(dx / abs(dx)) + move.start.get_x()
                    y_trajectory = move.start.get_y()
                    if (
                        board.get_cell(x_trajectory, y_trajectory).get_piece()
                        is not None
                    ):
                        return False
                # Along Y-axis
                for i in range(1, abs(dy)):
                    x_trajectory = move.start.get_x()
                    y_trajectory = i * int(dy / abs(dy)) + move.start.get_y()
                    if (
                        board.get_cell(x_trajectory, y_trajectory).get_piece()
                        is not None
                    ):
                        return False
                return True

            # Queen going in diagonal
            elif abs(dx) == abs(dy):
                for i in range(1, abs(dx)):
                    x_trajectory = i * int(dx / abs(dx)) + move.start.get_x()
                    y_trajectory = i * int(dy / abs(dy)) + move.start.get_y()
                    if (
                        board.get_cell(x_trajectory, y_trajectory).get_piece()
                        is not None
                    ):
                        return False
                return True
        else:
            return False

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """
        possible_moves = []

        # Diagonal 1
        nx = x - 1
        ny = y - 1
        while nx >= 0 and ny >= 0:
            possible_moves.append((nx, ny))
            nx -= 1
            ny -= 1

        # Diagonal 2
        nx = x - 1
        ny = y + 1
        while nx >= 0 and ny <= 7:
            possible_moves.append((nx, ny))
            nx -= 1
            ny += 1

        # Diagonal 3
        nx = x + 1
        ny = y - 1
        while nx <= 7 and ny >= 0:
            possible_moves.append((nx, ny))
            nx += 1
            ny -= 1

        # Diagonal 4
        nx = x + 1
        ny = y + 1
        while nx <= 7 and ny <= 7:
            possible_moves.append((nx, ny))
            nx += 1
            ny += 1

        # X-axis left
        nx = x - 1
        while nx >= 0:
            possible_moves.append((nx, y))
            nx -= 1

        # X-axis right
        nx = x + 1
        while nx <= 7:
            possible_moves.append((nx, y))
            nx += 1

        # Y-axis down
        ny = y - 1
        while ny >= 0:
            possible_moves.append((x, ny))
            ny -= 1

        # Y-axis top
        ny = y + 1
        while ny <= 7:
            possible_moves.append((x, ny))
            ny += 1

        return possible_moves

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  Q  "


class King(Piece):
    """Base class for the king pieces

    Implements the properties, attributes and functions specific to kings.

    Attributes
    ----------
    white : bool
        Whether the piece is white or black.
    x : int
        x coordinate of piece on board
    y : int
        y coordinate of piece on board
    killed: bool
        Whether the piece has been killed by opponent or not. Initialized to False
    has_moved: bool
        Whether the piece has already moved during a game or not. Initialized to False
    castling_done: bool
        Whether the piece has already realized castling. Initialized to False.
    """

    type = "king"

    def __init__(self, *args, **kwargs):
        """Initialization of the king.

        Parameters
        ----------
        white : bool
            Whether the piece is white or black.
        x : int
            initial x coordinate of piece on board
        y : int
            initial y coordinate of piece on board

        """
        super().__init__(*args, **kwargs)
        self.castling_done = False
        self.has_moved = False

    def piece_deepcopy(self):
        """Method to create an uncorrelated clone of the piece.

        Returns
        -------
        King
            Exact copy of self.
        """
        copied_piece = King(self.white, self.x, self.y)
        copied_piece.killed = self.killed
        copied_piece.castling_done = self.castling_done
        copied_piece.has_moved = self.has_moved
        return copied_piece

    def set_castling_done(self, castling_done):
        self.castling_done = castling_done

    def piece_move_authorized(self, start, end):
        """Method to verify if is a move is authorized in terms of movements.

        Parameters
        ----------
        start: engine.Cell
            Starting cell for movement check (current cell).
        end: engine.Cell
            Landing cell for movement check.

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """

        if not super().piece_move_authorized(start=start, end=end):
            return False
        dx = end.get_x() - start.get_x()
        dy = end.get_y() - start.get_y()

        if abs(dx) < 2 and abs(dy) < 2:
            return True
        else:
            return False

    def can_move(self, board, move):
        """Method to verify if a move is authorized within a board.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to and on which the movement is tested
        move: engine.Move
            Move object to be tested

        Returns
        -------
        bool
            Whether the movement is authorized by the piece possibilities or not.
        """
        # Checking if movement is authorized
        authorized_move = self.piece_move_authorized(move.start, move.end)
        if authorized_move:
            # Verifying that the landing cell is not threatened by some adversary material
            if move.end.is_threatened(board, self.is_white()):
                return False
            else:
                return True
        # If move is not authorized it could mean that player is trying to do a special move, i.e. castling
        else:
            # Checking castling conditions on the right then on the left
            if (
                not self.castling_done
                and not self.has_moved
                and (move.end.y == 6 or move.end.y == 2)
            ):
                if move.end.y == 6:  # Roque vers la droite
                    # Getting the rook for castling
                    rook_to_move = board.get_cell(move.start.x, 7).get_piece()
                    rook_starting_coordinates = (move.start.x, 7)
                    rook_ending_coordinates = (move.start.x, 5)

                    # Listing cells that must not have material on
                    if isinstance(rook_to_move, Rook):
                        must_be_empty_cells = [
                            board.get_cell(move.start.x, 5),
                            board.get_cell(move.start.x, 6),
                        ]
                        must_not_be_threatened_cells = [
                            board.get_cell(move.start.x, 4),
                            board.get_cell(move.start.x, 5),
                            board.get_cell(move.start.x, 6),
                        ]
                    else:
                        return False

                elif move.end.y == 2:  # Roque vers la gauche
                    rook_to_move = board.get_cell(move.start.x, 0).get_piece()
                    rook_starting_coordinates = (move.start.x, 0)
                    rook_ending_coordinates = (move.start.x, 3)

                    # Getting the rook
                    if isinstance(rook_to_move, Rook):
                        # Listing cells that must not have material on
                        must_be_empty_cells = [
                            board.get_cell(move.start.x, 1),
                            board.get_cell(move.start.x, 2),
                            board.get_cell(move.start.x, 3),
                        ]
                        must_not_be_threatened_cells = [
                            board.get_cell(move.start.x, 2),
                            board.get_cell(move.start.x, 3),
                            board.get_cell(move.start.x, 4),
                        ]
                    else:
                        return False

                else:
                    return False

                # Verifying conditions for listed cells
                empty_cells_check = True
                not_threatened_cells = True
                for cll in must_be_empty_cells:
                    if cll.get_piece() is not None:
                        empty_cells_check = False
                for cll in must_not_be_threatened_cells:
                    if cll.is_threatened(board, self.is_white()):
                        not_threatened_cells = False

                # Verify that all conditions are met and completes the move so that it has the full castling information
                # to operate all the movements
                conditions_to_castling = [
                    not rook_to_move.has_moved,
                    empty_cells_check,
                    not_threatened_cells,
                ]
                if all(conditions_to_castling):
                    move.complementary_castling = (
                        rook_to_move,
                        board.get_cell(
                            rook_starting_coordinates[0], rook_starting_coordinates[1]
                        ),
                        board.get_cell(
                            rook_ending_coordinates[0], rook_ending_coordinates[1]
                        ),
                    )
                    return True
                else:
                    return False
            return False

    def get_potential_moves(self, x, y):
        """Method to list all the possible moves from coordinates. Only uses authorized movements, no other pieces on a
        board.

        Parameters
        ----------
        x: int
            x coordinate of the piece
        y: int
            y coordinate of the piece

        Returns
        -------
        list
            List of authorized moves
        """
        possible_moves = []

        # All possible moves
        combos = [(1, 0), (1, 1), (-1, 1), (1, -1), (-1, -1), (0, 1), (0, -1), (-1, 0)]
        for nx, ny in combos:
            if 0 <= x + nx <= 7 and 0 <= y + ny <= 7:
                possible_moves.append((nx + x, ny + y))

        # Add castling as potential moves if not done yet
        if not self.has_moved:
            possible_moves.append((x, 1))
            possible_moves.append((x, 6))

        return possible_moves

    def get_str(self):
        """Method to represent the piece as a string.

        Returns
        -------
        str
            String representation of the piece
        """
        return "  K  "

    def is_checked(self, board):
        """Method to verify that the king at its current position is not threatened / checked by opponent material.

        Parameters
        ----------
        board: engine.Board
            Board to which the piece belongs to

        Returns
        -------
        bool
            Whether the king is checked or not.
        """

        return board.get_cell(self.x, self.y).is_threatened(board, self.white)

    # def is_checked_mate(self, board):
    #     if not self.is_checked(board):
    #         return False
    #
    #     for i in range(8):
    #         for j in range(8):
    #             piece = board.get_cell(i, j).get_piece()
    #             if piece is not None:
    #                 if piece.is_white() == self.is_white():
    #                     for move in piece.get_potential_moves(piece.x, piece.y):
    #                         selected_move = Move(None, board.get_cell(i, j),
    #                                              board.get_cell(move[0], move[1]))
    #                         verified_move = piece.can_move(board, selected_move)
    #
    #                         if verified_move:
    #                             copied_board = board.copy()



import copy
import pickle
import numpy as np

from pyalapin.player.player import Player
import pyalapin.engine.material as material
import pyalapin.engine.move as move


class EasyAIPlayer(Player):
    """
    AI Player class with simple rules and alpha/beta pruning to speed up research of the best move in the future.

    Attributes
    ----------
    is_white_side : bool
        Whether the player plays with white or black Pieces.
    piece_weights: dict
        Values of the different pieces.
    pieces_positions_weights: dict
        Values for each piece to be on a certain position.
    random_coeff: int
        Coefficient of randomness that will be added to the move score.
    """

    piece_weights = {
        "pawn": 10,
        "knight": 30,
        "bishop": 30,
        "rook": 50,
        "queen": 90,
        "king": 900,
    }
    piece_positions_weights = {
        "pawn": [
            [0.0] * 8,
            [5.0] * 8,
            [1.0, 1.0, 2.0, 3.0, 3.0, 2.0, 1.0, 1.0],
            [0.5, 0.5, 1.0, 2.5, 2.5, 1.0, 0.5, 0.5],
            [0.0, 0.0, 0.0, 2.0, 2.0, 0.0, 0.0, 0.0],
            [0.5, -0.5, -1.0, 0.0, 0.0, -1.0, -0.5, 0.5],
            [0.5, 1.0, 1.0, -2.0, -2.0, 1.0, 1.0, 0.5],
            [0.0] * 8,
        ],
        "bishop": [
            [-2.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -2.0],
            [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
            [-1.0, 0.0, 0.5, 1.0, 1.0, 0.5, 0.0, -1.0],
            [-1.0, 0.5, 0.5, 1.0, 1.0, 0.5, 0.5, -1.0],
            [-1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, -1.0],
            [-1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, -1.0],
            [-1.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.5, -1.0],
            [-2.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -2.0],
        ],
        "knight": [
            [-5.0, -4.0, -3.0, -3.0, -3.0, -3.0, -4.0, -5.0],
            [-4.0, -2.0, 0.0, 0.0, 0.0, 0.0, -2.0, -4.0],
            [-3.0, 0.0, 1.0, 1.5, 1.5, 1.5, 1.0, 0.0, -3.0],
            [-3.0, 0.5, 1.5, 2.0, 2.0, 1.5, 0.5, -3.0],
            [-3.0, 0.0, 1.5, 2.0, 2.0, 1.5, 0.0, -3.0],
            [-3.0, 0.5, 1.0, 1.5, 1.5, 1.0, 0.5, -3.0],
            [-4.0, -2.0, 0.0, 0.5, 0.5, 0.0, -2.0, -4.0],
            [-5.0, -4.0, -3.0, -3.0, -3.0, -3.0, -4.0, -5.0],
        ],
        "rook": [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5],
            [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
            [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
            [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
            [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
            [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
            [0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0],
        ],
        "queen": [
            [-2.0, -1.0, -1.0, -0.5, -0.5, -1.0, -1.0, -2.0],
            [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
            [-1.0, 0.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.0, -1.0],
            [-0.5, 0.0, 0.5, 0.5, 0.5, 0.5, 0.0, -0.5],
            [0.0, 0.0, 0.5, 0.5, 0.5, 0.5, 0.0, 0.0],
            [-1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, -1.0],
            [-1.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
            [-2.0, -1.0, -1.0, -0.5, -0.5, -1.0, -1.0, -2.0],
        ],
        "king": [
            [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
            [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
            [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
            [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
            [-2.0, -3.0, -3.0, -4.0, -4.0, -3.0, -3.0, -2.0],
            [-1.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0, -1.0],
            [2.0, 2.0, 0.0, 0.0, 0.0, 0.0, 2.0, 2.0],
            [2.0, 3.0, 1.0, 0.0, 0.0, 1.0, 3.0, 2.0],
        ],
    }

    def __init__(self, white_side, random_coeff=0, *args, **kwargs):
        """Initialization of the player.

        Parameters
        ----------
        white_side : bool
            Whether the player plays with white or black pieces.
        random_coeff: int
            Coefficient of randomness that will be added to the move score.
        """
        super().__init__(white_side=white_side, *args, **kwargs)
        self.color = "white" if self.white_side else "black"
        self.random_coeff = random_coeff
        if self.white_side:
            # Reverse position values for white pieces player
            for key, values in self.piece_positions_weights.items():
                new_values = []
                for i in range(len(values)):
                    new_values.append(values[-i])
                self.piece_positions_weights[key] = new_values

    def __str__(self):
        """Initialization of the player.

        Returns
        -------
        str
            String representation of the player
        """
        return "EasyAIPlayer"

    def _get_possible_moves(self, board, is_white=None):
        """Initialization of the player.

        Parameters
        ----------
        board: Board
            Board on which to look for the possible moves.
        is_white : bool or None
            If we want the possible moves for a different player, can be used.
        """
        if is_white is None:
            is_white = self.white_side()

        if self.white_side != is_white:

            class TempPlayer:
                def __init__(self):
                    self.white_side = is_white

                def is_white_side(self):
                    return self.white_side

            player = TempPlayer()
        else:
            player = self
        color = {True: "white", False: "black"}[is_white]

        possible_moves = []
        # Iterate of pieces
        for type_piece in board.all_material[color]["alive"].keys():
            for piece in board.all_material[color]["alive"][type_piece]:
                # Get potential moves as coordinates
                piece_available_moves = piece.get_potential_moves(piece.x, piece.y)

                # Iterate over piece possible moves
                for mv in piece_available_moves:
                    # Verify that the move is actually possible
                    selected_move = move.Move(
                        player,
                        board,
                        board.get_cell(piece.x, piece.y),
                        board.get_cell(mv[0], mv[1]),
                    )
                    # Keep only of possible
                    # Test letting this test in _alpha_beta
                    if selected_move.is_possible_move(check_chess=False):
                        possible_moves.append(selected_move)

        return possible_moves

    def _select_move_from_score(self, moves, method="max"):
        """Method to select a move according to the score on the board after the move.
        Maximum or minimum score can be selected.

        Parameters
        ----------
        moves : list
            list of moves among which to chose the best one.
        method: str in {"min", "max"}
            Whether to select the move leading to minimum or maximum score

        Returns
        -------
        move.Move
            Selected move
        float
            Score corresponding to the selected move
        """
        # Compute scores
        scores = {}
        for mv in moves:
            mv_ = mv.deepcopy()
            # mv_ = pickle.loads(pickle.dumps(mv, -1))
            # mv_ = copy.deepcopy(mv)
            mv_.move_pieces()
            score = self._score_board(mv_.board)
            scores.append(score)

        # Selection os scores
        if method == "max":
            all_indexes = np.where(scores == np.max(scores))[0]
        elif method == "min":
            all_indexes = np.where(scores == np.min(scores))[0]
        else:
            raise ValueError("MIN OR MAX ALGO, selected %s" % method)
        # If several moves lead to the same score, select randomly
        if len(all_indexes) > 1:
            perm = np.random.permutation(len(all_indexes))[0]
            final_index = all_indexes[perm]

        # Return
        return moves[int(final_index)], scores[final_index]

    # def _def_get_best_score(self, moves, method="max"):
    #     all_scores = []
    #     for mv in moves:
    #         mv_ = copy.deepcopy(mv)
    #         mv_.move_pieces()
    #         score = self._score_board(mv_.board)
    #         all_scores.append(score)
    #
    #     if method == "max":
    #         return np.max(all_scores)
    #     elif method == "min":
    #         return np.min(all_scores)
    #     else:
    #         raise ValueError

    def _search_tree(self, init_board, depth=2, method="max"):
        possible_moves = self._get_possible_moves(init_board)
        if depth == 1:
            best_move, best_score = self._select_move_from_score(
                possible_moves, method=method
            )
            return best_move, best_score
        else:
            new_method = {"max": "min", "min": "max"}
            scores = []
            for p_mv in possible_moves:
                p_mv = p_mv.deepcopy()
                # p_mv = pickle.loads(pickle.dumps(p_mv, -1))
                # p_mv = copy.deepcopy(p_mv)
                p_mv.move_pieces()
                _, score = self._search_tree(
                    p_mv.board, depth=depth - 1, method=new_method[method]
                )
                scores.append(score)

            if method == "max":
                best_score = np.max(scores)
            else:
                best_score = np.min(score)

            best_indexes = np.where(np.array(scores) == best_score)[0]
            final_index = best_indexes[np.random.permutation(len(best_indexes))[0]]
            return possible_moves[int(final_index)], best_score

    def _score_move(self, move):
        all_scores = {}
        mv_ = move.deepcopy()
        # mv_ = pickle.loads(pickle.dumps(move, -1))
        # mv_ = copy.deepcopy(move)
        mv_.move_pieces()
        score = self._score_board(mv_.board)

        return move, score

    def _alpha_beta(
        self,
        init_board,
        init_move=None,
        depth=2,
        alpha=-10000,
        beta=10000,
        is_white=None,
        draw_board=False,
    ):
        """Method to reccursively look for the best move. Implements alpha-beta pruning to test most possible moves.

        Parameters
        ----------
        init_board : engine.Board
            Current state of the game as board
        init_move: move.Move or None
            Not sure yet
        depth: int
            How many turns to look into the future (also used for recursion).
        alpha: int
            Max score value for alpha pruning
        beta: int
            Max score value for beta pruning
        is_white: None or bool
            Can be used if we want to use the method for other color than self
        draw_board: bool
            Whether or not to draw the board in terminal

        Returns
        -------
        int
            score of "best" selected move
        move
            "best" selected move
        """
        # If want to use other color than self.
        if is_white is None:
            is_white = self.white_side

        if draw_board:
            init_board.draw()

        # End of recursion
        if depth == 0:
            # Score current board and return it
            score = self._score_board(init_board)
            return score, init_move
        elif is_white == self.is_white_side():
            # Get moves
            possible_moves = self._get_possible_moves(init_board, is_white=is_white)
            # Iterate over moves and keep the best one.
            best_score = -10000
            best_move = None
            i = 0
            for p_mv in possible_moves:
                i += 1
                p_mv_ = p_mv.deepcopy()
                p_mv_.move_pieces()
                # Get best move if p_mv is actually made
                score, _ = self._alpha_beta(
                    p_mv_.board,
                    init_move=p_mv_,
                    depth=depth - 1,
                    alpha=alpha,
                    beta=beta,
                    is_white=not is_white,
                )
                if self.random_coeff > 0:
                    random_noise = np.random.randint(0, self.random_coeff)
                else:
                    random_noise = 0
                best_move = [best_move, p_mv][
                    np.argmax([best_score, score + random_noise])
                ]
                best_score = np.max([best_score, score + random_noise])

                if best_score >= beta:
                    return best_score, best_move
                alpha = np.max((alpha, best_score))

            return best_score, best_move

        else:
            possible_moves = self._get_possible_moves(init_board, is_white=is_white)

            best_score = 10000
            best_move = None
            for p_mv in possible_moves:
                p_mv_ = p_mv.deepcopy()
                p_mv_.move_pieces()
                score, _ = self._alpha_beta(
                    p_mv_.board,
                    init_move=p_mv_,
                    depth=depth - 1,
                    alpha=alpha,
                    beta=beta,
                    is_white=is_white,
                )

                best_move = [best_move, p_mv][np.argmin([best_score, score])]
                best_score = np.min([best_score, score])

                if best_score <= alpha:
                    return best_score, best_move
                beta = np.min([beta, best_score])
            return best_score, best_move

    def random_move(self, board):
        for i in np.random.permutation(8):
            for j in np.random.permutation(8):
                if board.get_cell(i, j).get_piece() is not None:
                    if (
                        board.get_cell(i, j).get_piece().is_white()
                        == self.is_white_side()
                    ):
                        selected_piece = board.get_cell(i, j).get_piece()
                        ###print('AI Selected Piece', selected_piece)
                        possible_moves = selected_piece.get_potential_moves(i, j)

                        verified_move = False
                        random_move = np.random.permutation(len(possible_moves))
                        index = 0
                        ###print('Verifying Moves,', len(possible_moves), 'Moves Possibles')
                        while not verified_move and index < len(random_move):
                            selected_move = possible_moves[random_move[index]]
                            selected_move = move.Move(
                                self,
                                board,
                                board.get_cell(i, j),
                                board.get_cell(selected_move[0], selected_move[1]),
                            )
                            verified_move = selected_move.is_possible_move()
                            index += 1

                        if verified_move:
                            ###print('Move is verified, ')
                            return selected_move
        ###print('No moved found, aborting...')

    def time_to_play(self, board, depth=3, draw_board=False):
        """Method that must be called to ask AI player to move.

        Parameters
        ----------
        board : engine.Board
            board on which to play
        depth: int
            Tree best move search depth
        draw_board: bool
            Whether or not to draw the board in terminal

        Returns
        -------
        move.Move
            Best move according to board and parameters
        """
        if draw_board:
            board.draw()
        # current_score = self._score_board(board)
        sel_score, sel_move = self._alpha_beta(board, depth=depth)

        return sel_move

    def _score_board(self, board):
        """Method to score a board according to player policy.

        Parameters
        ----------
        board : engine.Board
            board to score

        Returns
        -------
        float
            Score corresponding to the board and the player's parameters
        """
        score = 0
        # Positive score for player pieces
        for piece_type in board.all_material[self.color]["alive"].keys():
            for piece in board.all_material[self.color]["alive"][piece_type]:
                score += self.piece_weights[piece_type]
                score += self.piece_positions_weights[piece_type][piece.x][piece.y]
        own_king = board.all_material[self.color]["alive"]["king"]
        if len(own_king) == 0:
            score -= 1000
        else:
            own_king = own_king[0]
            if board.get_cell(own_king.x, own_king.y).is_threatened(
                board, own_king.is_white()
            ):
                score -= 1000

        adv_color = "white" if self.color == "black" else "black"
        # Negative score for opponent pieces.
        for piece_type in board.all_material[adv_color]["alive"].keys():
            for piece in board.all_material[adv_color]["alive"][piece_type]:
                score -= self.piece_weights[piece_type]
                score -= self.piece_positions_weights[piece_type][piece.x][piece.y]
        adv_king = board.all_material[adv_color]["alive"]["king"]

        if len(adv_king) == 0:
            score -= 1000
        else:
            adv_king = adv_king[0]
            if board.get_cell(adv_king.x, adv_king.y).is_threatened(
                board, adv_king.is_white()
            ):
                score += 1000
        return score



import numpy as np

from pyalapin.engine.move import Move


class Player:
    """Base class players. If human player only here to ensure that right player is playing right Pieces.
    For AI-based players can implement more operations.

    Attributes
    ----------
    is_white_side : bool
        Whether the player plays with white or black Pieces.

    """

    def __init__(self, white_side):
        """Initialization of the player.

        Parameters
        ----------
        white_side : bool
            Whether the player plays with white or black pieces.
        """
        self.white_side = white_side
        self.random_number = np.random.randint(0, 1000, 1)

    def __str__(self):
        """Creates a string representation of the player.

        Returns
        -------
        str
            String representation of the player

        """
        return "NormalPlayer%i" % self.random_number

    def is_white_side(self):
        """Method to access the player's side.

        Returns
        -------
        bool
            color of pieces played by player.

        """
        return self.white_side

    def time_to_play(self, board):
        """Potential method that returns a move.


        Parameters
        ----------
        board : engine.Board on which to play.

        Returns
        -------
        move.Move move to operate on board.

        """
        pass


class AIRandomPlayer(Player):
    """
    A first AI that plays totally randomly. Selects one move among all possibles and plays it.
    """

    def __str__(self):
        """Creates a string representation of the player.

        Returns
        -------
        str
            String representation of the player
        """

        return "AIRandomPlayer"

    def time_to_play(self, board):
        """Potential method that returns a move.


        Parameters
        ----------
        board : engine.Board on which to play.

        Returns
        -------
        move.Move move to operate on board.
        """

        # Random selection of Piece to play
        for i in np.random.permutation(8):
            for j in np.random.permutation(8):
                # Verfies there is a Piece to play.
                if board.get_cell(i, j).get_piece() is not None:
                    if (
                        board.get_cell(i, j).get_piece().is_white()
                        == self.is_white_side()
                    ):
                        # Get potential moves
                        selected_piece = board.get_cell(i, j).get_piece()
                        print("AI Selected Piece", selected_piece)
                        possible_moves = selected_piece.get_potential_moves(i, j)

                        # We can must verify that a move is playable
                        verified_move = False
                        # Random selection of move
                        random_move = np.random.permutation(len(possible_moves))
                        index = 0
                        print(
                            "Verifying Moves,", len(possible_moves), "Moves Possibles"
                        )
                        # Stop only with a move that can be played.
                        while not verified_move and index < len(random_move):
                            # Select and check move.
                            selected_move = possible_moves[random_move[index]]
                            selected_move = Move(
                                self,
                                board,
                                board.get_cell(i, j),
                                board.get_cell(selected_move[0], selected_move[1]),
                            )
                            verified_move = selected_move.is_possible_move()
                            index += 1

                        # If move is ok, break loop and return it
                        if verified_move:
                            print("Move is verified, ")
                            return selected_move
        print("No moved found, aborting...")



import os

import numpy as np

# import tensorflow as tf

from pyalapin.player.player import Player
import pyalapin.engine.material as material
import pyalapin.engine.move as move


class Memory(object):
    def __init__(self, max_memory=200):
        self.max_memory = max_memory
        self.memory = list()

    def remember(self, m):
        self.memory.append(m)
        self.memory = self.memory[max(len(self.memory) - self.max_memory, 0) :]

    def random_access(self):
        rn = np.random.randint(0, max(len(self.memory), 1))
        return self.memory[rn]


class MyPlayer(Player):
    def __init__(self, path_to_model="", epsilon_explorer=0.15, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.color = "white" if self.white_side else "black"
        self.path_to_model = path_to_model
        self.epsilon_explorer = epsilon_explorer

        in_ = tf.keras.layers.Input(shape=(8, 8, 6))
        out = tf.keras.layers.Conv2D(16, kernel_size=3, padding="SAME")(in_)
        out = tf.keras.layers.Conv2D(32, kernel_size=3, padding="SAME")(out)
        out = tf.keras.layers.Conv2D(16, kernel_size=3, padding="SAME", strides=2)(out)
        out = tf.keras.layers.Conv2D(8, kernel_size=3, padding="SAME", strides=2)(out)
        out = tf.keras.layers.Flatten()(out)
        out = tf.keras.layers.Dense(8)(out)
        out = tf.keras.layers.Dense(1)(out)
        if path_to_model == "":
            self.path_to_model = "./"
            for _ in range(10):
                self.path_to_model += str(np.random.randint(10))

        if not os.path.exists(path_to_model):
            self.model = tf.keras.Model(inputs=in_, outputs=out)
            self.model.compile(loss="mse", optimizer="Adam")
        else:
            print("Loading Model")
            self.model = tf.keras.models.load_model(path_to_model)
            self.model.compile(loss="mse", optimizer="Adam")
            print("Model Loaded")
        print(self.model.summary())
        self.memory = Memory()

    def _score_board(self, board, white_side=None):
        if white_side == None:
            white_side = self.white_side

        one_hot_encode_board = board.one_hot_encode()

        if self.white_side != white_side:
            one_hot_encode_board = np.flip(one_hot_encode_board)

        score = self.model.predict(
            np.expand_dims(one_hot_encode_board, 0).astype("float32")
        )
        return score[0][0]

    def _get_possible_moves(self, board, is_white=None):
        if is_white is None:
            is_white = self.white_side()

        color = {True: "white", False: "black"}[is_white]
        possible_moves = []
        for type_piece in board.all_material[color]["alive"].keys():
            for piece in board.all_material[color]["alive"][type_piece]:
                piece_available_moves = piece.get_potential_moves(piece.x, piece.y)
                for mv in piece_available_moves:
                    selected_move = move.Move(
                        self,
                        board,
                        board.get_cell(piece.x, piece.y),
                        board.get_cell(mv[0], mv[1]),
                    )
                    if selected_move.is_possible_move():
                        possible_moves.append(selected_move)
        return possible_moves

    def _select_move_from_score(self, moves, train=True):
        if train:
            do_random_move = np.random.randint(100) <= self.epsilon_explorer * 100
            if do_random_move:
                print("RANDOM MOVE SELECTED")
                index_random_move = np.random.randint(len(moves))
                random_move = moves[index_random_move]
                mv_ = random_move.deepcopy()
                mv_.move_pieces()
                random_score = self._score_board(mv_.board)
                return random_move, random_score

        all_scores = {}
        for mv in moves:
            mv_ = mv.deepcopy()
            mv_.move_pieces()
            score = self._score_board(mv_.board)
            all_scores[mv] = score

        scores = list(all_scores.values())
        all_indexes = np.where(scores == np.max(scores))[0]

        perm = np.random.permutation(len(all_indexes))[0]
        final_index = all_indexes[perm]
        return list(all_scores.keys())[int(final_index)], scores[final_index]

    def _base_scoring(self, board, game_is_won="", white_side=None):
        if white_side is None:
            white_side = self.white_side
        piece_weights = {
            "pawn": 10,
            "knight": 30,
            "bishop": 30,
            "rook": 50,
            "queen": 90,
            "king": 900,
        }
        piece_positions_weights = {
            "pawn": [
                [0.0] * 8,
                [5.0] * 8,
                [1.0, 1.0, 2.0, 3.0, 3.0, 2.0, 1.0, 1.0],
                [0.5, 0.5, 1.0, 2.5, 2.5, 1.0, 0.5, 0.5],
                [0.0, 0.0, 0.0, 2.0, 2.0, 0.0, 0.0, 0.0],
                [0.5, -0.5, -1.0, 0.0, 0.0, -1.0, -0.5, 0.5],
                [0.5, 1.0, 1.0, -2.0, -2.0, 1.0, 1.0, 0.5],
                [0.0] * 8,
            ],
            "bishop": [
                [-2.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -2.0],
                [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
                [-1.0, 0.0, 0.5, 1.0, 1.0, 0.5, 0.0, -1.0],
                [-1.0, 0.5, 0.5, 1.0, 1.0, 0.5, 0.5, -1.0],
                [-1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, -1.0],
                [-1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, -1.0],
                [-1.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.5, -1.0],
                [-2.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -2.0],
            ],
            "knight": [
                [-5.0, -4.0, -3.0, -3.0, -3.0, -3.0, -4.0, -5.0],
                [-4.0, -2.0, 0.0, 0.0, 0.0, 0.0, -2.0, -4.0],
                [-3.0, 0.0, 1.0, 1.5, 1.5, 1.5, 1.0, 0.0, -3.0],
                [-3.0, 0.5, 1.5, 2.0, 2.0, 1.5, 0.5, -3.0],
                [-3.0, 0.0, 1.5, 2.0, 2.0, 1.5, 0.0, -3.0],
                [-3.0, 0.5, 1.0, 1.5, 1.5, 1.0, 0.5, -3.0],
                [-4.0, -2.0, 0.0, 0.5, 0.5, 0.0, -2.0, -4.0],
                [-5.0, -4.0, -3.0, -3.0, -3.0, -3.0, -4.0, -5.0],
            ],
            "rook": [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5],
                [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
                [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
                [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
                [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
                [-0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.5],
                [0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0],
            ],
            "queen": [
                [-2.0, -1.0, -1.0, -0.5, -0.5, -1.0, -1.0, -2.0],
                [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
                [-1.0, 0.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.0, -1.0],
                [-0.5, 0.0, 0.5, 0.5, 0.5, 0.5, 0.0, -0.5],
                [0.0, 0.0, 0.5, 0.5, 0.5, 0.5, 0.0, 0.0],
                [-1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, -1.0],
                [-1.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0],
                [-2.0, -1.0, -1.0, -0.5, -0.5, -1.0, -1.0, -2.0],
            ],
            "king": [
                [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
                [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
                [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
                [-3.0, -4.0, -4.0, -5.0, -5.0, -4.0, -4.0, -3.0],
                [-2.0, -3.0, -3.0, -4.0, -4.0, -3.0, -3.0, -2.0],
                [-1.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0, -1.0],
                [2.0, 2.0, 0.0, 0.0, 0.0, 0.0, 2.0, 2.0],
                [2.0, 3.0, 1.0, 0.0, 0.0, 1.0, 3.0, 2.0],
            ],
        }

        score = 0
        for piece_type in board.all_material["white"]["alive"].keys():
            for piece in board.all_material["white"]["alive"][piece_type]:
                score += piece_weights[piece_type]
                score += np.flip(piece_positions_weights[piece_type])[piece.x][piece.y]

        for piece_type in board.all_material["black"]["alive"].keys():
            for piece in board.all_material["black"]["alive"][piece_type]:
                score -= piece_weights[piece_type]
                score -= piece_positions_weights[piece_type][piece.x][piece.y]
        if not white_side:
            score = -1 * score
        if game_is_won:
            score += 1000
        elif not game_is_won:
            score -= 1000

        return score

    def _reinforce(self, board, best_next_score, batch_size=12, white_side=None):
        if white_side is None:
            white_side = self.white_side
        score = self._base_scoring(board, white_side=white_side)
        print("SCORE during reinforce phase", score)
        self.memory.remember((board.one_hot_encode(white_side=white_side), score))

        input_states = np.zeros((batch_size, 8, 8, 6))
        target_q = np.zeros((batch_size, 1))

        for i in range(batch_size):
            (s_i, r_i) = self.memory.random_access()
            input_states[i] = s_i
            target_q[i] = r_i
        self.model.train_on_batch(input_states, target_q)
        self.model.save(self.path_to_model)

    def __str__(self):
        return "MyAIPlayer"

    def time_to_play(self, board, white_side=None):
        if white_side is None:
            white_side = self.is_white_side()
        possible_moves = self._get_possible_moves(board, is_white=white_side)
        best_move, best_score = self._select_move_from_score(possible_moves)

        self._reinforce(board, best_score, white_side=white_side)
        return best_move



import numpy as np
import os

from kivy.app import App
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.textinput import TextInput
from kivy.uix.button import Button
from kivy.uix.popup import Popup

from kivy.graphics import Rectangle, Color, Canvas

from pyalapin.engine.engine import ChessGame


class LoginScreen(GridLayout):
    def __init__(self, **kwargs):
        """
        Base class for a Login Screen, not used yet
        """
        super(LoginScreen, self).__init__(**kwargs)
        self.cols = 8
        self.add_widget(Label(text=""))
        self.username = TextInput(multiline=False)
        self.add_widget(self.username)
        self.add_widget(Label(text="password"))
        self.password = TextInput(password=True, multiline=False)
        self.add_widget(self.password)

        self.add_widget(Label(text="password2"))
        self.password2 = TextInput(password=True, multiline=False)
        self.add_widget(self.password2)

        self.add_widget(Rectangle(pos=(10, 10), size=(500, 500)))


class DisplayableCell(Button):
    """Base class to represent a Cell as Button"""

    def __init__(self, row, column, **kwargs):
        """
        Initialization of the representation of the cell.

        Parameters
        ----------

        row: int
            row coordinate of the Cell
        column: int
            column coordinate of the Cell
        """
        super(DisplayableCell, self).__init__(**kwargs)
        self.row = row
        self.column = column


class BoardInterface(GridLayout):
    """
    Main class to represent and display the board, as well as to play a chess game.

    Attributes
    ----------
    path_to_illustrations: str
        Path to the images to use to display cells & pieces
    game: engine.ChessGame
        game to actually represent
    ai_playing: bool
        whether or not an AI is playing (only one of the players for now)
    cols: int
        number of columns of the board
    rows: int
        number of rows of the board
    to_play_player: player.Player
        player who should play next
    first_cell_clicked: engince.Cell or None
        First part of a piece movement, used to memorize the first cell selected by user who would like to move a piece.
    cells: list of DisplayableCells
        List of cells constituting the board
    """

    def __init__(self, game, **kwargs):
        """
        Initialization of the board display.

        Parameters
        ----------
        game: engine.Game to represent

        """
        super(BoardInterface, self).__init__(**kwargs)
        self.path_to_illustrations = "illustrations"
        self.game = game

        if game.ai:
            self.ai_playing = True
        else:
            self.ai_playing = False

        self.cols = 8
        self.rows = 8

        self.to_play_player = game.player1

        self.first_cell_clicked = None

        self.cells = []

        # Initialization of the display of the board
        for i in range(self.rows):
            line = []
            for j in range(self.cols):
                # Alternate cells color for better board perception
                if (i % 2 == 0 and j % 2 == 0) or (i % 2 == 1 and j % 2 == 1):
                    color = (0.4, 0.4, 0.8, 1)
                    c_img = "b"
                else:
                    color = (0.4, 0.8, 0.4, 1)
                    c_img = "w"
                # self.add_widget(Button(text='Button %i %i' % (i, j), background_color=color,
                #                 background_down='illustrations/white_pawn.png',
                #                 background_normal='illustrations/white_queen.png'))
                piece = game.board.get_cell(i, j).get_piece()

                if piece is not None:
                    path_to_img = c_img

                    if piece.is_white():
                        piece_color = (1, 1, 1, 1)  # For text color, could be removed
                        path_to_img += "w"
                    else:
                        piece_color = (0, 0, 0, 1)
                        path_to_img += "b"
                    path_to_img += "_" + piece.get_str().replace(" ", "") + ".png"
                    path_to_down_img = "down_" + path_to_img

                    path_to_img = os.path.join(self.path_to_illustrations, path_to_img)
                    path_to_down_img = os.path.join(
                        self.path_to_illustrations, path_to_down_img
                    )

                    piece = piece.get_str()
                    button = DisplayableCell(
                        # text=piece, # Remove it for prettier results :)
                        on_press=self.click_cell,
                        row=i,
                        column=j,
                        color=piece_color,
                        background_normal=path_to_img,
                        border=(0, 0, 0, 0),
                        background_down=path_to_down_img,
                    )
                else:
                    # No piece to display
                    piece = ""
                    piece_color = (1, 1, 1, 1)  # For text color could be removed
                    path_to_img = c_img + ".png"
                    # Unclicked
                    path_to_down_img = "down_" + path_to_img

                    path_to_img = os.path.join(self.path_to_illustrations, path_to_img)
                    path_to_down_img = os.path.join(
                        self.path_to_illustrations, path_to_down_img
                    )

                    button = DisplayableCell(
                        # text=piece, # Remove for prettier results :)
                        background_normal=path_to_img,
                        on_press=self.click_cell,
                        row=i,
                        column=j,
                        color=piece_color,
                        border=(0, 0, 0, 0),
                        background_down=path_to_down_img,
                    )
                self.add_widget(button)
                line.append(button)
            self.cells.append(line)

    def reset_game(self, button):
        """
        Method used to reset a game when clicked on the reset button

        Parameters
        ----------
        button: Button
            button used to click for reset
        """
        print("On click, Reset", button)
        self.game.reset_game()
        self.update()

    def update(self):
        """
        Method used to update the display of the board. Actually redraws everythin from start.
        """
        board = self.game.board
        for i in range(self.rows):
            for j in range(self.cols):
                if (i % 2 == 0 and j % 2 == 0) or (i % 2 == 1 and j % 2 == 1):
                    c_img = "b"
                else:
                    c_img = "w"

                piece = board.get_cell(i, j).get_piece()
                if piece is not None:
                    path_to_img = c_img
                    if piece.is_white():
                        piece_color = (1, 1, 1, 1)
                        path_to_img += "w"
                    else:
                        piece_color = (0, 0, 0, 1)
                        path_to_img += "b"
                    piece = piece.get_str()
                    path_to_img += "_" + piece.replace(" ", "") + ".png"
                else:
                    piece = ""
                    piece_color = (1, 1, 1, 1)
                    path_to_img = c_img + ".png"

                path_to_down_img = "down_" + path_to_img
                path_to_img = os.path.join(self.path_to_illustrations, path_to_img)
                path_to_down_img = os.path.join(
                    self.path_to_illustrations, path_to_down_img
                )
                # self.cells[i][j].text = piece
                # self.cells[i][j].color = piece_color
                self.cells[i][j].background_normal = path_to_img
                self.cells[i][j].background_down = path_to_down_img

    def finish_game(self, winner):
        """
        Method used to trigger a display of the end of a game.

        Parameters
        ----------
        winner: player.Player
            Winner of self.Game if the game is finished.
        """
        popup = Popup(title="Game finished", auto_dismiss=False)
        popup.bind(on_dismiss=self.reset_game)

        box = BoxLayout()
        box.add_widget(Label(text="Congratulations %s has won" % winner))
        restart_button = Button(text="Restart a game!")
        restart_button.bind(on_press=popup.dismiss)
        box.add_widget(restart_button)

        popup.bind(on_dismiss=self.reset_game)
        popup.content = box
        popup.open()

    def click_cell(self, event):
        """
        Main trigger when a cell is clicked.
        Works in a serie of two clicks:
            - First click is to define start cell (or cell of the piece we want to move)
            - Second click is to define the landing cell (or cell we want to move the piece to)

        Parameters
        ----------
        event: Event
            Click on a Board's cell event triggering the method.
        """
        # Reverse the backgrounds when a cell is clicked, to indicate it has been clicked.
        (
            self.cells[event.row][event.column].background_normal,
            self.cells[event.row][event.column].background_down,
        ) = (
            self.cells[event.row][event.column].background_down,
            self.cells[event.row][event.column].background_normal,
        )
        # If no previous cell has been clicked, then it's the start cell that has been clicked,
        # In this case it is store, waiting fot the click on the landinc cell.
        if self.first_cell_clicked is None:
            self.first_cell_clicked = (event.row, event.column)
        # In this case the player has clicked twice on the same cell.
        # It is considered as a cancellation of the first click
        elif (
            self.first_cell_clicked[0] == event.row
            and self.first_cell_clicked[1] == event.column
        ):
            print("Selection Aborted")
            self.first_cell_clicked = None
        # In this cas, actually move the piece.
        else:
            start_x = self.first_cell_clicked[0]
            start_y = self.first_cell_clicked[1]
            end_x = event.row
            end_y = event.column

            # Try and move if possible the piece
            validated_move, winner = self.game.move_from_coordinates(
                self.game.to_play_player, start_x, start_y, end_x, end_y
            )

            # In case move is ok
            if validated_move:
                self.update()

                # If AI is playing, then trigger its next move with time_to_play method.
                if self.ai_playing:
                    print("Time for AI")

                    # Resets background colors of player
                    ai_move = self.game.player2.time_to_play(self.game.board)
                    self.game.board.draw()
                    game_is_on = self.game.move(ai_move, self.game.player2)

                    # Verify game is still going on
                    # Actually we consider that an AI cannot trigger an impossible move here
                    # Maybe should be modified ?

                    if game_is_on[0]:
                        self.update()
                        (
                            self.cells[ai_move.start.x][
                                ai_move.start.y
                            ].background_normal,
                            self.cells[ai_move.start.x][
                                ai_move.start.y
                            ].background_down,
                        ) = (
                            self.cells[ai_move.start.x][
                                ai_move.start.y
                            ].background_down,
                            self.cells[ai_move.start.x][
                                ai_move.start.y
                            ].background_normal,
                        )
                        (
                            self.cells[ai_move.end.x][ai_move.end.y].background_normal,
                            self.cells[ai_move.end.x][ai_move.end.y].background_down,
                        ) = (
                            self.cells[ai_move.end.x][ai_move.end.y].background_down,
                            self.cells[ai_move.end.x][ai_move.end.y].background_normal,
                        )
                    else:
                        if isinstance(game_is_on[1], str):
                            self.finish_game(game_is_on[1])
                        else:
                            # Can we be here ?
                            pass

            # Verify is there is or not a winner and if game is finished.
            elif isinstance(winner, str):
                print("WINNER", winner)
                self.finish_game(winner)
                return None

            # In this case, game was not possible, reset last clicks so that the player can restart
            # and redefine its move.
            else:
                popup = Popup(
                    title="Unable Move",
                    content=Label(
                        text="Your selected move is not possible, please, select another one."
                    ),
                    size_hint=(None, None),
                    size=(15, 15),
                )
                popup.open()

            if not self.ai_playing:
                # Resets values befor next move
                # If AI is playing, it is handled with self.update()
                row, col = self.first_cell_clicked
                (
                    self.cells[row][col].background_normal,
                    self.cells[row][col].background_down,
                ) = (
                    self.cells[row][col].background_down,
                    self.cells[row][col].background_normal,
                )
                (
                    self.cells[event.row][event.column].background_normal,
                    self.cells[event.row][event.column].background_down,
                ) = (
                    self.cells[event.row][event.column].background_down,
                    self.cells[event.row][event.column].background_normal,
                )

            self.first_cell_clicked = None
            if not validated_move:
                self.update()


class ChessApp(App):
    """
    Main app to use to play game, by calling MyApp().buil() and then player.
    """

    def __init__(self, play_with_ai=False, **kwargs):
        """
        Initialization, with precision whether or not playing with AI.
        """
        super().__init__(**kwargs)
        self.play_with_ai = play_with_ai

    def build(self):
        """
        Builds the game and the display board along with it.
        """
        game = ChessGame(automatic_draw=False, ai=self.play_with_ai)
        print("game created")
        return BoardInterface(game)
